import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from sys import platform
import json
import ntpath
import unicodedata
import threading
import os
import logging
from ftplib import FTP
from ftplib import FTP_TLS
from threading import Lock
from shutil import copy2
from datetime import datetime, timedelta

# --- constant connection values
ftpServerName = ""
ftpU = ""
ftpP = ""
directoriesToWatch = []
useTLS = True
dictOfWatchedDir = {}
synchAtStartupBool = False
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)
lock = Lock()
workingDir = "/"
startBackupTime = "02:00"
backupDurationInHours = 1
delayToSend = -1
listOfFilesToBeSent = []
respectBackupHours = False


def sendDataCheck():
    # we want to avoid sending data within an interval that goes among <backupDurationInHours> hour before the backup time and <backupDurationInHours> hour after the backup time
    try:
        if (respectBackupHours):
            if (startBackupTime!="" and (":" in startBackupTime)):
                listOfParam = startBackupTime.split(":")
                backupHour = listOfParam[0]
                backupMinute = listOfParam[1]

                minorCheck = datetime.today() - timedelta(hours=backupDurationInHours, minutes=0)
                majorCheck = datetime.today() + timedelta(hours=backupDurationInHours, minutes=0)

                date = datetime.now()
                dateAndTimeOfBackup = date.replace(hour=int(backupHour), minute=int(backupMinute))

                if (dateAndTimeOfBackup > minorCheck and dateAndTimeOfBackup < majorCheck):
                    return False
                else:
                    return True
            else:
                #if anything goes wrong we return true
                return True
        else:
            return True

    except Exception as inst:
        #if anything goes wrong we return true
        return True

class Watcher(threading.Thread):
    # DIRECTORY_TO_WATCH = "/home/teo/Desktop/temp"

    def __init__(self, localDirPath):
        try:
            threading.Thread.__init__(self)
            self.observer = Observer()
            self.localDirPath = localDirPath
        except Exception as inst:
            logger.error("__init__ in Watcher Error for " + self.localDirPath + " - " + timeStamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly

    def run(self):
        try:
            event_handler = Handler()
            self.observer.schedule(event_handler, self.localDirPath, recursive=True)
            self.observer.start()
            logger.debug("The thread to monitor " + self.localDirPath + " was run")

            try:
                while True:
                    time.sleep(5)
            except:
                self.observer.stop()
                logger.error("An exception occured in the thread loop of " + self.localDirPath)
                # we want a service that is always run
                w = Watcher(self.localDirPath)
                logger.info("Restart the thread to observe changes in files in" + self.localDirPath)
                w.start()

            logger.info("Join the thread for " + self.localDirPath)
            self.observer.join()

        except Exception as inst:
            logger.error("run in Watcher Error for " + self.localDirPath + " - " + timeStamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly


class Handler(FileSystemEventHandler):

    @staticmethod
    def on_any_event(event):
        result = None
        try:
            if event.is_directory:
                # do nothing
                result = None
            elif event.event_type == 'created':
                # Take any action here when a file is first created.
                # when the file is create the "modified" event is also generated
                # so we do not anything
                logger.info("Received created event - %s." % event.src_path)
                # if (event.src_path in dictOfWatchedDir):
                #    moveFTPFiles(ftpServerName, ftpU, ftpP, event.src_path, dictOfWatchedDir[event.src_path], useTLS)

            elif event.event_type == 'modified':
                fileToUpload = event.src_path
                # Taken any action here when a file is modified.
                logger.info("Received modified event - %s." % fileToUpload)

                path_without_filename = path_without_leaf(fileToUpload)
                if (path_without_filename in dictOfWatchedDir):
                    if (delayToSend<=0 and sendDataCheck()):
                        # directly send the file without any delay
                        moveFTPFiles(ftpServerName, ftpU, ftpP, fileToUpload, dictOfWatchedDir[path_without_filename], useTLS)
                    else:
                        # insert the name of the file in a list that will be elaborated only once every delayToSend seconds
                        if (fileToUpload not in listOfFilesToBeSent):
                            listOfFilesToBeSent.append(fileToUpload)
            elif event.event_type == 'moved':
                # on linux, when a file is modified, the system create a temporal file (.goutputstream...)
                #  and them move it on the right one
                fileToUpload = event.dest_path

                # Taken any action here when a file is modified.
                logger.info("Received moved event - %s. - " % fileToUpload)

                path_without_filename = path_without_leaf(fileToUpload)
                if (path_without_filename in dictOfWatchedDir):
                    if (delayToSend<=0 and sendDataCheck()):
                        # directly send the file without any delay
                        moveFTPFiles(ftpServerName, ftpU, ftpP, fileToUpload, dictOfWatchedDir[path_without_filename], useTLS)
                    else:
                        # insert the name of the file in a list that will be elaborated only once every delayToSend seconds
                        if (fileToUpload not in listOfFilesToBeSent):
                            listOfFilesToBeSent.append(fileToUpload)
                # path_without_filename = path_without_leaf(fileToUpload)
                # if (path_without_filename in dictOfWatchedDir):
                #     moveFTPFiles(ftpServerName, ftpU, ftpP, fileToUpload, dictOfWatchedDir[path_without_filename], useTLS)
            elif event.event_type == 'deleted':
                # do nothing
                result = None
        except Exception as inst:
            logger.error("on_any_event Error - " + timeStamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly


def path_without_leaf(path):
    result = ""
    try:
        head, tail = ntpath.split(path)
        result = head
    except Exception as inst:
        logger.error("path_without_leaf Error - " + timeStamp())
        logger.error(type(inst))  # the exception instance
        logger.error(inst.args)  # arguments stored in .args
        logger.error(inst)  # __str__ allows args to be printed directly
        result = ""
    return result

#return the name of the file + extension
def path_leaf(path):
    result = ""
    try:
        head, tail = ntpath.split(path)
        result = tail or ntpath.basename(head)
    except Exception as inst:
        logger.error("Path_leaf Error - " + timeStamp())
        logger.error(type(inst))  # the exception instance
        logger.error(inst.args)  # arguments stored in .args
        logger.error(inst)  # __str__ allows args to be printed directly
        result = ""
    return result


def moveFTPFiles(serverName, userName, passWord, fileToUpload, remoteDirectoryPath, useTLS=False):
    lock.acquire(True)
    """Connect to an FTP server and bring down files to a local directory"""
    if (serverName != '' and userName != '' and passWord != ''):
        try:
            ftp = None
            if useTLS:
                ftp = FTP_TLS(serverName, timeout=120)
                if (ftp == None):
                    logger.info("LOG moveFTPFiles 3TLS ftp null")
            else:
                ftp = FTP(serverName)
            ftp.login(userName, passWord)
            if useTLS:
                ftp.prot_p()
            ftp.cwd(remoteDirectoryPath)
            ftp.set_pasv(True)
            filesMoved = 0

            try:

                logger.info("Connecting...")

                # create a full local filepath
                localFileName = path_leaf(fileToUpload)
                if localFileName != "" and not localFileName.startswith("."):

                    #create a copy of the file before sending it
                    tempFile = create_temporary_copy(fileToUpload, workingDir, localFileName)

                    # open a the local file
                    fileObj = open(tempFile, 'rb')
                    # Download the file a chunk at a time using RETR
                    ftp.storbinary('STOR ' + localFileName, fileObj)
                    # Close the file
                    fileObj.close()
                    filesMoved += 1
                    # remove the temp file
                    # to remove it I need to have the right permissions
                    os.chmod(tempFile, 777)
                    os.remove(tempFile)

                logger.info("Uploaded file: " + str(fileToUpload) + " on " + timeStamp())
            except Exception as inst:

                logger.error(type(inst))  # the exception instance
                logger.error(inst.args)  # arguments stored in .args
                logger.error(inst)  # __str__ allows args to be printed directly
                logger.error("Connection Error - " + timeStamp())
            ftp.quit()  # Close FTP connection
            ftp = None
        except Exception as inst:
            logger.error("Couldn't find server")
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly
    else:
        logger.error(
            "Connection was not possible because one of the following var was not set in the configuration file: ftpServerName , ftpU or ftpP")

    lock.release()


def create_temporary_copy(path, temp_dir, fileName):
    if platform == "win32":
        fileToCheck = temp_dir + '\\' + fileName
    else:
        fileToCheck = temp_dir + '/' + fileName
    # if the file already exists, we will delete it
    exists = os.path.isfile(fileToCheck)
    if exists:
        os.chmod(fileToCheck, 777)
        os.remove(fileToCheck)
    temp_path = os.path.join(temp_dir, fileName)
    copy2(path, temp_path)
    return temp_path


def timeStamp():
    """returns a formatted current time/date"""
    import time
    return str(time.strftime("%a %d %b %Y %I:%M:%S %p"))


def synchAtStartup():
    try:
        if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
            for dest in directoriesToWatch:
                if ('localDirPath' in dest and 'remoteDirPath' in dest):
                    logger.info("synchAtStartup: synchronizing " + dest['localDirPath'])
                    arr = os.listdir(dest['localDirPath'])
                    for file in arr:
                        if os.path.isdir(file) != True:
                            moveFTPFiles(ftpServerName, ftpU, ftpP, dest['localDirPath'] + "/" + file,
                                         dest['remoteDirPath'], useTLS)
    except Exception as inst:
        logger.error("synchAtStartup Error - " + timeStamp())
        logger.error(type(inst))  # the exception instance
        logger.error(inst.args)  # arguments stored in .args
        logger.error(inst)  # __str__ allows args to be printed directly

def elaborateAllChangedFiles():
    global listOfFilesToBeSent
    if (sendDataCheck()):
        while listOfFilesToBeSent:
            fileToUpload = listOfFilesToBeSent[0]
            ##fileToUpload = listOfFilesToBeSent.pop(0)
            path_without_filename = path_without_leaf(fileToUpload)
            if (path_without_filename in dictOfWatchedDir):
                moveFTPFiles(ftpServerName, ftpU, ftpP, fileToUpload, dictOfWatchedDir[path_without_filename], useTLS)
                #when the system has finished to upload the file, we remove it from the list
                listOfFilesToBeSent.remove(listOfFilesToBeSent[0])
            #listOfFilesToBeSent.pop(-1)
    else:
        logger.info("Not sending data because there could be a running update")
    #re-instantiate the thread that should send data
    threading.Timer(delayToSend, elaborateAllChangedFiles).start()


if __name__ == '__main__':
    conf = {
      "configuration":{
        "directoriesToWatch":[
          {
            "localDirPath": "/home/user/temp",
            "remoteDirPath": "temp1"
          },
          {
            "localDirPath": "/home/user/temp2",
            "remoteDirPath": "temp2"
          }
        ],
        "synchAtStartup":True,
        "ftpServerName":"server.domain.com",
        "ftpUser":"ftp-user",
        "ftpPass":"ftpPass",
        "useTLS":True,
        "workingDir": "/home/ftpWorkingDir",
        "delayToSend": 20,
        "startBackupTime": "10:12",
        "backupDurationInHours": 1,
        "respectBackupHours": True
        }
    }
    try:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.DEBUG)

        # create a file handler
        handler = logging.FileHandler('log.log')
        handler.setLevel(logging.DEBUG)

        # create a logging format
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)

        # add the handlers to the logger
        logger.addHandler(handler)

        logger.info("loading configuration set")
        # read configuration parameters from the config.json file
        conf_path ='config.json'
        if os.path.exists(conf_path):
            with open(conf_path) as json_data_file:
                data = json.load(json_data_file)
        if 'configuration' in data:
            conf = data['configuration']
        if 'directoriesToWatch' in conf:
            directoriesToWatch = conf['directoriesToWatch']
        if 'ftpServerName' in conf:
            ftpServerName = conf['ftpServerName']
        if 'ftpUser' in conf:
            ftpU = conf['ftpUser']
        if 'ftpPass' in conf:
            ftpP = conf['ftpPass']
        if 'useTLS' in conf:
            useTLS = conf['useTLS']
        if 'synchAtStartup' in conf:
            synchAtStartupBool = conf['synchAtStartup']
        if 'workingDir' in conf:
            workingDir = conf['workingDir']
        if 'startBackupTime' in conf:
            startBackupTime = conf['startBackupTime']
        if 'backupDurationInHours' in conf:
            backupDurationInHours = conf['backupDurationInHours']
        if 'respectBackupHours' in conf:
            respectBackupHours = conf['respectBackupHours']
        if (not respectBackupHours):
            logger.info("The script will NOT respect backup hours (so it will upload file also during backup hours)")

        #check if the folder exists, otherwise, create it
        try:
            if not os.path.exists(workingDir):
                os.makedirs(workingDir)
        except Exception as inst:
            if platform == "win32":
                workingDir = '/'
            else:
                workingDir = 'C:\\'
        try:
            if 'delayToSend' in conf:
                if (isinstance(conf['delayToSend'], int)):
                    delayToSend = conf['delayToSend']
                else:
                    delayToSend = int(conf['delayToSend'],10)
        except Exception as inst:
            delayToSend = -1

        wrongPath = False

        logger.info("check correctness of links")
        if platform == "win32":
            # Windows...
            if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
                for dest in directoriesToWatch:
                    if ('/' in dest['localDirPath']):
                        wrongPath = True
                        logger.error("In the configuraton file you set a path (" + dest[
                            'localDirPath'] + ") as a linux-like path but you are under linux")
            #same check for the workingDir variable
            if ('/' in workingDir):
                wrongPath = True
                logger.error("In the configuraton file you set a path (" + workingDir + ") as a linux-like path but you are under linux")
        elif platform == "linux" or platform == "linux2" or platform == "darwin":
            # linux or OSX
            if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
                for dest in directoriesToWatch:
                    if ('\\' in dest['localDirPath']):
                        wrongPath = True
                        logger.error("In the configuraton file you set a path (" + dest[
                            'localDirPath'] + ") as a windows-like path but you are under linux")

            # same check for the workingDir variable
            if ('\\' in workingDir):
                wrongPath = True
                logger.error(
                    "In the configuraton file you set a path (" + workingDir + ") as a windows-like path but you are under linux")
        if (wrongPath):
            logger.error("Exiting due to errors in the path (e.g., we are under Windows but the user set a linux path)")
            exit()
        else:

            if synchAtStartupBool:
                logger.info("synchronizing files at startup")
                synchAtStartup()
            if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
                for dest in directoriesToWatch:
                    if 'localDirPath' in dest and 'remoteDirPath' in dest:
                        localDirPath = unicodedata.normalize('NFKD', dest['localDirPath']).encode('ascii', 'ignore')
                        remoteDirPath = unicodedata.normalize('NFKD', dest['remoteDirPath']).encode('ascii',
                                                                                                        'ignore')
                        dictOfWatchedDir[localDirPath] = remoteDirPath

                        w = Watcher(localDirPath)
                        logger.info("Start the thread to observe changes in files in" + localDirPath)
                        w.start()

            if (delayToSend>0):
                threading.Timer(delayToSend, elaborateAllChangedFiles).start()
    except Exception as inst:
        logger.error("main Error - " + timeStamp())
        logger.error(type(inst))  # the exception instance
        logger.error(inst.args)  # arguments stored in .args
        logger.error(inst)  # __str__ allows args to be printed directly


import re
import json
import sys
import time
import re
import urllib.parse
import clipboard


def fetch_to_requests(fetch):
    raw = fetch.strip().split('fetch(', 1)[-1][:-2]
    url, data = re.fullmatch(r'"(.*)", {(.*)', raw, re.DOTALL).groups()
    url_d = urllib.parse.urlsplit(url)
    print(url_d.netloc)
    print(url_d.fragment)
    print(url_d.scheme)
    print(url_d.hostname)
    url_format = "{}_{}".format(
        url_d.hostname.replace('.', '_'),
        re.sub(r'[^a-zA-Z0-9]', '_', url_d.path)
    )
    print(url_format)
    data = json.loads(f'{{{data}')

    headers = data['headers']
    headers["Host"] = url_d.netloc
    headers["Referer"] = data.get('referrer') or url
    headers["Origin"] = headers["Referer"][:-1] if headers["Referer"].endswith('/') else headers["Referer"]
    headers["Connection"] = "keep-alive"
    for k in [
        "Sec-Fetch-Dest",
        "Sec-Fetch-Mode",
        "Sec-Fetch-Site",
    ]:
        if (kl := k.lower()) in headers:
            headers[k] = headers.pop(kl)
    # headers["Sec-Fetch-Dest"] = "script",
    # headers["Sec-Fetch-Mode"] = "no-cors",
    # headers["Sec-Fetch-Site"] = "same-origin",
    # headers["Accept-Encoding"] =  "gzip, deflate, br"
    # headers["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,i"
    # "mage/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
    # headers["Accept-Language"] = "zh,zh-CN;q=0.9,en;q=0.8",
    headers_str = '{{\n        "User-Agent": get_ua(), {}    }}'.format(json.dumps(
        headers, separators=(",", ": "), indent=8)[1:-1])
    proxies_host = '127.0.0.1'
    proxies_port = 8389
    proxies = {
        "http": f"socks5://{proxies_host}:{proxies_port}",
        "https": f"socks5://{proxies_host}:{proxies_port}",
    }

    code_res = f'''import random
import requests
import pyquery
import logging
import os
import wrapt
import traceback
import functools

logger_debug = logging.debug
logger_info = logging.info
logger_warn = logging.warning
logger_error = logging.error

USE_PROXY = True
# USE_PROXY = False
USE_CACHE = True
# USE_CACHE = False
PASS_EX = True
# PASS_EX = False
TMP_PATH_CACHE = 'tmp_cache_html'


def save_cache_to_file(if_read_cache_key=None):
    """加上装饰器就会保存函数结果到文件，装饰器设置if_read_cache_key字符床，函数定义时设置对应名称参数，并在调用时设置True可从文件读cache"""
    if if_read_cache_key and not os.path.exists(TMP_PATH_CACHE):
        os.makedirs(TMP_PATH_CACHE)

    @wrapt.decorator
    def wrapper(func, _instance, args, kwargs):
        cache_path = os.path.join(TMP_PATH_CACHE, func.__name__)
        if isinstance(if_read_cache_key, str) and kwargs.get(if_read_cache_key) and os.path.exists(cache_path):
            with open(cache_path, 'rb') as f:
                result = f.read()
            logger_info('[{{}}]函数读取到缓存: {{}} | {{}} | {{}}'.format(
                func.__name__, os.path.realpath(cache_path)[:-len(cache_path)], cache_path, repr(result)[:30]))
        else:
            result = func(*args, **kwargs)
            if result:
                with open(cache_path, 'wb') as f:
                    f.write(result)
                    logger_debug('[{{}}]函数结果缓存到: {{}} | {{}} |'.format(
                        func.__name__, os.path.realpath(cache_path)[:-len(cache_path)], cache_path))
        return result

    return wrapper


def pass_ex(if_ex_return=None, exs=None):
    """加上装饰器就会捕获所有异常，异常返回if_ex_return，exs中的异常会对应打印error，不再exs里面则打印堆栈信息"""
    if not exs:
        exs = {{
            requests.exceptions.ConnectionError: '[ConnectionError] 代理失效或URL失效',
            requests.exceptions.ConnectTimeout: '[ConnectTimeout] 连接握手超时',
            requests.exceptions.ReadTimeout: '[ReadTimeout] 读取数据过程中超时',
        }}

    @wrapt.decorator
    def wrapper(func, _instance, args, kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as ex:
            if type(ex) in exs:
                logger_warn('{{}} {{}} 将返回  {{}}'.format(exs[type(ex)], func.__name__, repr(if_ex_return)))
            else:
                logger_warn("[{{}}] {{}} {{}}\\n{{}}".format(type(ex), ex, func.__name__, traceback.format_exc()))
            return if_ex_return

    return wrapper


def get_proxies():
    if not USE_PROXY:
        return None
    proxies_host = '{proxies_host}'
    proxies_port = {proxies_port}
    proxies = {{
        "http": f"socks5://{{proxies_host}}:{{proxies_port}}",
        "https": f"socks5://{{proxies_host}}:{{proxies_port}}",
    }}
    return proxies


def get_ua():
    return random.choice([
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.131 Safari/537.36"])


@save_cache_to_file(if_read_cache_key='_use_cache')
def get_html_{url_format}(url=r"{url}", _use_cache=True):
    headers = {headers_str}
    proxies = get_proxies()
    res = requests.{data['method'].lower()}(
        url=url,
        headers=headers,
        verify=False,
        # stream=True,
        stream=False,
        proxies=proxies,
        timeout=125,
    )
    result = res.content
    logger_info('{{}}采集到{{}}  ->  {{}}'.format(f"使用代理[{{proxies.get('http')}}]" if proxies else "不使用代理", url, repr(result)[:50]))
    return result


get_html_{url_format} = pass_ex(if_ex_return='')(get_html_{url_format})


def task(url):
    pq_html = pyquery.PyQuery(get_html_{url_format}(url=url, _use_cache=USE_CACHE) or '<></>')('html')
    
    for item in pq_html:
        logger_debug(item.text)


if __name__ == '__main__':
    logging.basicConfig(
        level=logging.INFO,
        format="[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S"
    )
    tmp_logger_debug = functools.partial(print, '[DEBUG]')
    logger_debug = lambda x: tmp_logger_debug(repr(x))
    task(r"{url}")
'''
    return code_res


def stdout_write(msg: str):
    sys.stdout.write(msg)
    sys.stdout.flush()


def stderr_write(msg: str):
    sys.stderr.write(msg)
    sys.stderr.flush()


# noinspection PyProtectedMember,PyUnusedLocal,PyIncorrectDocstring,DuplicatedCode
def nb_print(*args, sep=' ', end='\n', file=None, flush=True):
    print_raw = print
    args = (str(arg) for arg in args)  # REMIND 防止是数字不能被join
    if file == sys.stderr:
        stderr_write(sep.join(args))  # 如 threading 模块第926行，打印线程错误，希望保持原始的红色错误方式，不希望转成蓝色。
    elif file in [sys.stdout, None]:
        # 获取被调用函数在被调用时所处代码行数
        line = sys._getframe().f_back.f_lineno
        # 获取被调用函数所在模块文件名
        file_name = sys._getframe(1).f_code.co_filename
        # sys.stdout.write(f'"{__file__}:{sys._getframe().f_lineno}"    {x}\n')
        stdout_write(
            f'\033[0;34m{time.strftime("%H:%M:%S")}  "{file_name}:{line}"   \033[0;30;44m{sep.join(args)}\033[0m{end} \033[0m')  # 36  93 96 94
    else:
        print_raw(*args, sep=sep, end=end, file=file)


def gen_from_clipboard():
    text = clipboard.paste()
    nb_print(f"{'*' * 80}\n{text}\n{'*' * 80}")
    if text and text.strip().startswith('fetch("http'):
        return fetch_to_requests(text)


if __name__ == '__main__':
    res = gen_from_clipboard()
    # python -m d22d.utils.f2r
    print(res)
    with open('tmp_spider.py', 'w', encoding='utf-8') as f:
        f.write(res)



import diskcache
import json
import os
import logging
import functools
import asyncio
import shutil

import typing

from d22d.model.midhardware import BaseStore
from d22d.utils.decorators import flyweight

logger = logging.getLogger('diskcachemodel')


class JSONDiskNoC(diskcache.Disk):
    def __init__(self, directory, compress_level=0, **kwargs):
        self.compress_level = compress_level
        super(JSONDiskNoC, self).__init__(directory, **kwargs)

    def put(self, key):
        json_bytes = json.dumps(key).encode('utf-8')
        return super(JSONDiskNoC, self).put(json_bytes)

    def get(self, key, raw):
        data = super(JSONDiskNoC, self).get(key, raw)
        return json.loads(data.decode('utf-8'))

    def store(self, value, read, key=None):
        if not read:
            value = json.dumps(value).encode('utf-8')
        return super(JSONDiskNoC, self).store(value, read)

    def fetch(self, mode, filename, value, read):
        data = super(JSONDiskNoC, self).fetch(mode, filename, value, read)
        if not read:
            data = json.loads(data.decode('utf-8'))
        return data


PATH_REL_DATA = 'data_disk_cache'

PATH_CWD = os.getcwd()
path_abs_base_add = functools.partial(os.path.join, PATH_CWD)
PATH_ABS_DATA = path_abs_base_add(PATH_REL_DATA)
path_abs_data_add = functools.partial(os.path.join, PATH_REL_DATA)
real_path_abs_data_add = functools.partial(os.path.join, PATH_ABS_DATA)


def get_realpath():
    return os.path.split(os.path.realpath(__file__))[0]


# sys.path.append(os.path.join(get_realpath(), './'))


def get_make_local_path(*path):
    local_path = os.path.join(*path)
    if not os.path.exists(local_path):
        os.makedirs(local_path)
    return local_path


@flyweight
class DiskCache:

    def __init__(self, path=PATH_REL_DATA):
        self.path = path
        self.cache = self.init_cache_default(path).stats()

    @staticmethod
    def init_cache_default(path):
        cache = diskcache.Cache(
            directory=path, disk=JSONDiskNoC, disk_compress_level=6,
            size_limit=int(1024 ** 4),  # 1TB
            cull_limit=0)
        cache.create_tag_index()
        return cache

    def get_cache(self, *args, **kwargs):
        return self.cache.get(*args, **kwargs)

    def set_cache(self, *args, **kwargs):
        with self.cache.transact():
            return self.cache.set(*args, **kwargs)

    def del_cache(self, key, *args, **kwargs):
        if key in self.cache:
            del self.cache[key]

    async def set_cache_async(self, key, value, expire=None, read=False, tag=None, retry=False):
        loop = asyncio.get_running_loop()
        future = loop.run_in_executor(None, self.cache.set, key, value, expire, read, tag, retry)
        result = await future
        return result

    def list_cache(self, keys=None, log=True):
        for idx, k in enumerate(self.cache.iterkeys()):
            if keys and k not in keys:
                continue
            else:
                if log:
                    if isinstance(log, int):
                        if not (idx % log):
                            print(f'[cache_print] {k}: {self.cache[k]}')
                    else:
                        print(f'[cache_print] {k}: {self.cache[k]}')
                yield k

    def cull_cache_default(self):
        """
        将缓存目录清理到只剩size_limit的大小
        """
        return self.cache.cull(retry=True)

    def free_self(self, raise_err=False):
        if os.path.exists(self.path):
            can_delete = True
            for root, fs, fns in os.walk(self.path):
                for fn in fns:
                    if not fn.startswith('cache.db'):
                        can_delete = False
                        if raise_err:
                            raise OSError(f"存在非缓存文件{os.path.join(root, fn)}")
                        else:
                            logger.warning(f"存在非缓存文件{os.path.join(root, fn)}")
            if can_delete:
                logger.info(f'删除文件夹：{self.path}:->{os.path.realpath(self.path)}')
                shutil.rmtree(self.path)


class DiskCacheStore(BaseStore):
    def __init__(self, location='/'):
        self.raw_location = location
        self.location = os.path.realpath(location)
        self.client = DiskCache(location)

    def count_data(self, data_type=None, *args, **kwargs):
        return self.client.cache.__len__()

    def list_data(self, data_type=None, *args, **kwargs):
        return self.client.list_cache()

    def check_data(self, position, data_type=None, *args, **kwargs):
        return position in self.client.cache

    def get_data(self, position: typing.Union[str], data_type=None, *args, **kwargs):
        return self.client.get_cache(position)

    def save_data(self, position, data, data_type=None, *args, **kwargs):
        return self.client.set_cache(position, data)

    def delete_data(self, position, data_type=None, *args, **kwargs):
        return self.client.del_cache(position)

    def get_position(self, position, data_type=None, *args, **kwargs):
        return os.path.join(self.location, position)

    def get_data_size(self, position, data_type=None, *args, **kwargs):
        return NotImplementedError

    def check_self(self, *args, **kwargs):
        return NotImplementedError

    def save_self(self, *args, **kwargs):
        return NotImplementedError

    def free_self(self, *args, **kwargs):
        return NotImplementedError


if __name__ == '__main__':
    a = DiskCache()
    b = DiskCache()
    c = DiskCache(path_abs_data_add("sas"))
    print(id(a))
    print(id(b))
    print(id(c))
    c.free_self()
    a.free_self()
    b.free_self()


# whipFTP, Copyrights Vishnu Shankar B,

import os
import re
import time
import logging
from os import listdir
from os.path import isfile, join
import shutil
import paramiko

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class ParamikoSftpClient(paramiko.SFTPClient):
    def cwd(self, path):
        self.chdir(path)

    def go_to_home(self, username):
        try:
            self.cwd('/home/' + username)
        except:
            self.cwd('/')


class SftpController:
    # /!\ Although the comments and variable names say 'file_name'/'file_anything' it inculdes folders also
    # Some functions in this class has no exception handling, it has to be done outside

    def __init__(self):
        # List to store file search and search keywords
        self.search_file_list = []
        self.detailed_search_file_list = []
        self.keyword_list = []

        # Variable to hold the max no character name in file list (used for padding in GUIs)
        self.max_len = 0
        self.max_len_name = ''

        # Variable to tell weather hidden files are enabled
        self.hidden_files = False

    def connect_to(self, Host, Username=' ', Password=' ', Port=22):
        self.transport = paramiko.Transport((Host, Port))
        self.transport.connect(username=Username, password=Password)
        self.ftp = ParamikoSftpClient.from_transport(self.transport)
        self.ftp.go_to_home(Username)

    def toggle_hidden_files(self):
        self.hidden_files = not self.hidden_files

    def get_detailed_file_list(self, ignore_hidden_files_flag=False):
        files = []
        for attr in self.ftp.listdir_attr():
            if (self.hidden_files is True or str(attr).split()[8][0] is not '.') or ignore_hidden_files_flag == True:
                files.append(str(attr))
        return files

    def get_file_list(self, detailed_file_list):
        self.max_len = 0
        self.max_len_name = ''
        file_list = []
        for x in detailed_file_list:
            # Remove details and append only the file name
            name = ' '.join(x.split()[8:])
            file_list.append(name)
            if (len(name) > self.max_len):
                self.max_len = len(name)
                self.max_len_name = name
        return file_list

    def get_detailed_search_file_list(self):
        return self.detailed_search_file_list

    def get_search_file_list(self):
        self.max_len = 0
        self.max_len_name = ''
        for name in self.search_file_list:
            if (len(name) > self.max_len):
                self.max_len = len(name)
                self.max_len_name = name
        return self.search_file_list

    def chmod(self, filename, permissions):
        self.ftp.chmod(filename, permissions)

    def is_there(self, path):
        try:
            self.ftp.stat(path)
            return True
        except:
            return False

    def rename_dir(self, rename_from, rename_to):
        self.ftp.rename(rename_from, rename_to)

    def move_dir(self, rename_from, rename_to, status_command, replace_command):
        if (self.is_there(rename_to) is True):
            if (replace_command(rename_from, 'File/Folder exists in destination folder') is True):
                self.delete_dir(rename_to, status_command)
            else:
                return
        try:
            self.ftp.rename(rename_from, rename_to)
            status_command(rename_from, 'Moved')
        except:
            status_command(rename_from, 'Failed to move')

    def copy_file(self, file_dir, copy_from, file_size, status_command, replace_command):
        # Change to script's directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        if not os.path.exists('copy_temps'):
            os.makedirs('copy_temps')
        os.chdir('copy_temps')
        # Save the current path so that we can copy later
        dir_path_to_copy = self.ftp.getcwd()
        # Change to the file's path and download it
        self.ftp.cwd(file_dir)
        self.download_file(copy_from, file_size, status_command, replace_command)
        # Change back to the saved path and upload it
        self.ftp.cwd(dir_path_to_copy)
        self.upload_file(copy_from, file_size, status_command, replace_command)
        # Delete the downloaded file
        os.remove(copy_from)
        status_command(copy_from, 'Deleted local file')

    def copy_dir(self, file_dir, copy_from, status_command, replace_command):
        # Change to script's directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)
        if not os.path.exists('copy_temps'):
            os.makedirs('copy_temps')
        os.chdir('copy_temps')
        # Save the current path so that we can copy later
        dir_path_to_copy = self.ftp.getcwd()
        # Change to the file's path and download it
        self.ftp.cwd(file_dir)
        self.download_dir(copy_from, status_command, replace_command)
        # Change back to the saved path and upload it
        self.ftp.cwd(dir_path_to_copy)
        self.upload_dir(copy_from, status_command, replace_command)
        # Delete the downloaded folder
        shutil.rmtree(copy_from)
        status_command(copy_from, 'Deleting local directory')

    def delete_file(self, file_name, status_command):
        try:
            self.ftp.remove(file_name)
            status_command(file_name, 'Deleted')
        except:
            status_command(file_name, 'Failed to delete')

    def delete_dir(self, dir_name, status_command):
        # Go into the directory
        self.ftp.cwd(dir_name)
        # Get file lists
        try:
            detailed_file_list = self.get_detailed_file_list(True)
        except:
            status_command(dir_name, 'Failed to delete directory')
            return
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If directory
            if (self.is_dir(file_details)):
                self.delete_dir(file_name, status_command)
            # If file
            else:
                self.delete_file(file_name, status_command)
        # Go back to parent directory and delete it
        try:
            self.ftp.cwd('..')
            status_command(dir_name, 'Deleting directory')
            self.ftp.rmdir(dir_name)
        except:
            status_command(dir_name, 'Failed to delete directory')

    def upload_file(self, file_name, file_size, status_command, replace_command):
        # Function to update status
        def upload_progress(transferred, remaining):
            status_command(file_name, str(min(round((transferred / file_size) * 100, 8), 100)) + '%')

        # Check if the file is already present in ftp server
        if (self.is_there(file_name)):
            if (replace_command(file_name, 'File exists in destination folder') is False):
                return
        # Try to upload file
        try:
            status_command(file_name, 'Uploading')
            self.ftp.put(file_name, file_name, callback=upload_progress)
            status_command(None, 'newline')
        except:
            status_command(file_name, 'Upload failed')
            return

    def upload_dir(self, dir_name, status_command, replace_command):
        # Change to directory
        os.chdir(dir_name)
        # Create directory in server and go inside
        try:
            if (not self.is_there(dir_name)):
                self.ftp.mkdir(dir_name)
                status_command(dir_name, 'Creating directory')
            else:
                status_command(dir_name, 'Directory exists')
            self.ftp.cwd(dir_name)
        except:
            status_command(dir_name, 'Failed to create directory')
            return
        # Cycle through items
        for filename in os.listdir():
            # If file upload
            if (isfile(filename)):
                self.upload_file(filename, os.path.getsize(filename), status_command, replace_command)
            # If directory, recursive upload it
            else:
                self.upload_dir(filename, status_command, replace_command)

        # Got to parent directory
        self.ftp.cwd('..')
        os.chdir('..')

    def download_file(self, ftp_file_name, file_size, status_command, replace_command):
        # Function to update progress
        def download_progress(transferred, remaining):
            status_command(ftp_file_name, str(min(round((transferred / file_size) * 100, 8), 100)) + '%')

        # Check if the file is already present in local directory
        if (isfile(ftp_file_name)):
            if (replace_command(ftp_file_name, 'File exists in destination folder') is False):
                return
        # Try to download file
        try:
            status_command(ftp_file_name, 'Downloading')
            self.ftp.get(ftp_file_name, ftp_file_name, callback=download_progress)
            status_command(None, 'newline')
        except:
            status_command(ftp_file_name, 'Download failed')

    def download_dir(self, ftp_dir_name, status_command, replace_command):
        # Create local directory
        try:
            if (not os.path.isdir(ftp_dir_name)):
                os.makedirs(ftp_dir_name)
                status_command(ftp_dir_name, 'Created local directory')
            else:
                status_command(ftp_dir_name, 'Local directory exists')
            os.chdir(ftp_dir_name)
        except:
            status_command(ftp_dir_name, 'Failed to create local directory')
            return
        # Go into the ftp directory
        self.ftp.cwd(ftp_dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list(True)
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If directory
            if (self.is_dir(file_details)):
                self.download_dir(file_name, status_command, replace_command)
            # If file
            else:
                self.download_file(file_name, int(self.get_properties(file_details)[3]), status_command,
                                   replace_command)
        # Got to parent directory
        self.ftp.cwd('..')
        os.chdir('..')

    def search(self, dir_name, status_command, search_file_name):
        # Go into the ftp directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list()
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            # If file_name matches the keyword, append it to search list
            if search_file_name.lower() in file_name.lower():
                if (self.ftp.getcwd() == '/'):
                    dir = ''
                else:
                    dir = self.ftp.getcwd()
                self.search_file_list.append(dir + '/' + file_name)
                self.detailed_search_file_list.append(file_details)
                status_command(dir + '/' + file_name, 'Found')
                # If directory, search it
            if (self.is_dir(file_details)):
                status_command(file_name, 'Searching directory')
                self.search(file_name, status_command, search_file_name)
        # Goto to parent directory
        self.ftp.cwd('..')

    def clear_search_list(self):
        del self.search_file_list[:]
        del self.detailed_search_file_list[:]

    def get_dir_size(self, dir_name):
        size = 0;
        # Go into the ftp directory
        self.ftp.cwd(dir_name)
        # Get file lists
        detailed_file_list = self.get_detailed_file_list()
        file_list = self.get_file_list(detailed_file_list)
        for file_name, file_details in zip(file_list, detailed_file_list):
            if (self.is_dir(file_details)):
                size += self.get_dir_size(file_name)
            else:
                size += int(self.get_properties(file_details)[3])
        # Goto to parent directory
        self.ftp.cwd('..')
        # return size
        return size

    def cwd_parent(self, name):
        if ('/' not in name): return name
        parent_name = '/'.join(name.split('/')[:-1])
        if (parent_name == ''): parent_name = '/'
        self.ftp.cwd(parent_name)
        return ''.join(name.split('/')[-1:])

    def mkd(self, name):
        self.ftp.mkdir(name)

    def pwd(self):
        return (self.ftp.getcwd())

    def get_properties(self, file_details):
        details_list = file_details.split()
        # Get file attributes
        file_attribs = details_list[0]
        # Get date modified
        date_modified = ' '.join(details_list[5:8])
        # Remove the path from the name
        file_name = ' '.join(details_list[8:])
        # Get size if it is not a directory
        if ('d' not in file_details[0]):
            file_size = details_list[4]
            return [file_name, file_attribs, date_modified, file_size]
        else:
            return [file_name, file_attribs, date_modified]

    def is_dir(self, file_details):
        return 'd' in file_details[0]

    def disconnect(self):
        if self.ftp:
            self.ftp.close()


class ParamikoFolderUploader(object):
    """
    paramoki 实现的文件夹上传
    """

    def __init__(self, host, port, user, password, local_dir: str, remote_dir: str,
                 path_pattern_exluded_tuple=('/.git/', '/.idea/', '/dist/', '/build/'),
                 file_suffix_tuple_exluded=('.pyc', '.log', '.gz'),
                 only_upload_within_the_last_modify_time=3650 * 24 * 60 * 60,
                 file_volume_limit=1000 * 1000, sftp_log_level=20):
        """
        :param host:
        :param port:
        :param user:
        :param password:
        :param local_dir:
        :param remote_dir:
        :param path_pattern_exluded_tuple: 命中了这些正则的直接排除
        :param file_suffix_tuple_exluded: 这些结尾的文件排除
        :param only_upload_within_the_last_modify_time: 仅仅上传最近多少天修改的文件
        :param file_volume_limit: 大于这个体积的不上传，单位b。
        """
        self._host = host
        self._port = port
        self._user = user
        self._password = password

        self._local_dir = str(local_dir).replace('\\', '/')
        if not self._local_dir.endswith('/'):
            self._local_dir += '/'
        self._remote_dir = str(remote_dir).replace('\\', '/')
        if not self._remote_dir.endswith('/'):
            self._remote_dir += '/'
        self._path_pattern_exluded_tuple = path_pattern_exluded_tuple
        self._file_suffix_tuple_exluded = file_suffix_tuple_exluded
        self._only_upload_within_the_last_modify_time = only_upload_within_the_last_modify_time
        self._file_volume_limit = file_volume_limit

        # noinspection PyTypeChecker
        t = paramiko.Transport((host, port))
        t.connect(username=user, password=password)
        self.sftp = paramiko.SFTPClient.from_transport(t)

        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(host, port=port, username=user, password=password, compress=True)
        self.ssh = ssh

    def _judge_need_filter_a_file(self, filename: str):
        ext = filename.split('.')[-1]
        if '.' + ext in self._file_suffix_tuple_exluded:
            return True
        for path_pattern_exluded in self._path_pattern_exluded_tuple:
            # print(path_pattern_exluded,filename)
            if re.search(path_pattern_exluded, filename):
                return True
        file_st_mtime = os.stat(filename).st_mtime
        volume = os.path.getsize(filename)
        if time.time() - file_st_mtime > self._only_upload_within_the_last_modify_time:
            return True
        if volume > self._file_volume_limit:
            return True
        return False

    def _make_dir(self, dirc, final_dir):
        """
        sftp.mkdir 不能直接越级创建深层级文件夹。
        :param dirc:
        :param final_dir:
        :return:
        """
        # print(dir,final_dir)
        try:
            self.sftp.mkdir(dirc)
            if dirc != final_dir:
                self._make_dir(final_dir, final_dir)
        except (FileNotFoundError,):
            parrent_dir = os.path.split(dirc)[0]
            self._make_dir(parrent_dir, final_dir)

    def upload(self):
        for parent, dirnames, filenames in os.walk(self._local_dir):
            for filename in filenames:
                file_full_name = os.path.join(parent, filename).replace('\\', '/')
                if not self._judge_need_filter_a_file(file_full_name):
                    remote_full_file_name = re.sub(f'^{self._local_dir}', self._remote_dir, file_full_name)
                    try:
                        logger.debug(f'正在上传文件，本地：{file_full_name}  --> 远程： {remote_full_file_name}')
                        self.sftp.put(file_full_name, remote_full_file_name)
                    except (FileNotFoundError,) as e:
                        # self.logger.warning(remote_full_file_name)
                        self._make_dir(os.path.split(remote_full_file_name)[0], os.path.split(remote_full_file_name)[0])
                        self.sftp.put(file_full_name, remote_full_file_name)
                else:
                    if '/.git' not in file_full_name and '.pyc' not in file_full_name:
                        logger.debug(f'根据过滤规则，不上传这个文件 {file_full_name}')



import zip_files.backend
import zip_files.click_extensions
import logging
import os
import pickle
import shutil
import io
import struct

from zipfile import (
    ZipFile, _EndRecData, BadZipFile, _ECD_SIZE,
    _ECD_OFFSET, _ECD_COMMENT, _ECD_LOCATION, _ECD_SIGNATURE,
    stringEndArchive64, sizeEndCentDir64, sizeEndCentDir64Locator, sizeCentralDir,
    structCentralDir, _CD_SIGNATURE, stringCentralDir, _CD_FILENAME_LENGTH,
    ZipInfo, _CD_EXTRA_FIELD_LENGTH, _CD_COMMENT_LENGTH, _CD_LOCAL_HEADER_OFFSET,
    MAX_EXTRACT_VERSION, ZipExtFile, _SharedFile, sizeFileHeader,
    structFileHeader, _FH_SIGNATURE, _FH_FILENAME_LENGTH, _FH_EXTRA_FIELD_LENGTH,
    _FH_GENERAL_PURPOSE_FLAG_BITS, stringFileHeader
)
from pathlib import Path

PICKLE_PATH = 'pkl_cache'

logger = logging.getLogger('ziputils')


def file_exists(path):
    return os.path.exists(path) and os.path.isfile(path)


def makedirs(path, check_dot=True):
    if path in ['/',  '']:
        return
    folder_name = path.split('\\')[-1].split('/')[-1]
    if check_dot:
        is_folder = '.' not in folder_name
        path = os.path.realpath(path)
        out_f_path = os.path.dirname(path)
        if is_folder:
            out_f_path = os.path.join(out_f_path, folder_name)
    else:
        out_f_path = os.path.realpath(path)
    if not os.path.exists(out_f_path):
        logger.info(f'正在创建新文件夹：{out_f_path}，因为{path}需要')
        os.makedirs(out_f_path)


makedirs(PICKLE_PATH)


def get_cache_pickle(key, default=None, refresh=False):
    path = f'{PICKLE_PATH}/{key}.pkl'
    if os.path.exists(path) and not refresh:
        try:
            with open(path, 'rb') as f:
                res = pickle.load(f)
            need_set = False
        except Exception as ex:
            print(f"pickle load: {type(ex)} {ex}")
            need_set = True
    else:
        need_set = True
    if need_set:
        if callable(default):
            default = default()
        set_cache_pickle(key, default)
        res = default
    return res


def set_cache_pickle(key, data):
    makedirs(PICKLE_PATH)
    with open(f'{PICKLE_PATH}/{key}.pkl', 'wb') as f:
        pickle.dump(data, f)


def remove_file(path):
    if os.path.exists(path):
        os.remove(path)


def remove_folder(path):
    if os.path.exists(path):
        shutil.rmtree(path)


def zip_folder(f_path, outfile, root_folder=None, auto_root=False, new_file=False):
    """

    :param f_path:
    :param outfile:
    :param root_folder: 创建一个新文件夹作为压缩包内的根目录
    :param auto_root: True 创建一个新文件夹作为作为压缩包内的根目录, 用outfile文件名来命名
    :param new_file: 每次删除老的zip，创建新zip包
    :return:
    """
    logger.info(f'正在压缩zip包{f_path} --> {outfile}')
    makedirs(outfile)

    if new_file:
        remove_file(outfile)

    files = Path(f_path).iterdir()
    if root_folder is None:
        root_folder = Path(f_path).name
        if auto_root:
            root_folder = Path(outfile).stem
    """Compress list of `files`.

    Args:
        root_folder (Path or None): folder name to prepend to `files` inside
            the zip archive
        exclude (list[str]): A list of glob patterns to exclude. Matching is
            done from the right on the path names inside the zip archive.
            Patterns must be relative (not start with a slash)
        exclude_from (list[str]): A list of filenames from which to read
            exclude patterns (one pattern per line)
        exclude_dotfiles (bool): If given as True, exclude all files starting
            with a dot.
        exclude_vcs (bool): If given as True, exclude files and directories
            used by common version control systems (Git, CVS, RCS, SCCS, SVN,
            Arch, Bazaar, Mercurial, and Darcs), e.g.  '.git/', '.gitignore'
            '.gitmodules' '.gitattributes' for Git
        exclude_git_ignores (bool): If given as True, exclude files listed in
            any '.gitignore' in the given `files` or its subfolders.
        outfile (Path): The path of the zip file to be written
        files (Iterable[Path]): The files to include in the zip archive
    """
    zip_files.backend.zip_files(
        debug=True,
        root_folder=root_folder,
        compression=8,
        exclude=[],
        exclude_from=[],
        exclude_dotfiles=False,
        exclude_vcs=False,
        exclude_git_ignores=False,
        outfile=outfile,
        files=files)
    logger.info(f'压缩完成：{f_path} --> {outfile}')


class GBKZipFile(ZipFile):

    def _RealGetContents(self):
        """Read in the table of contents for the ZIP file."""
        fp = self.fp
        try:
            endrec = _EndRecData(fp)
        except OSError:
            raise BadZipFile("File is not a zip file")
        if not endrec:
            raise BadZipFile("File is not a zip file")
        if self.debug > 1:
            print(endrec)
        size_cd = endrec[_ECD_SIZE]  # bytes in central directory
        offset_cd = endrec[_ECD_OFFSET]  # offset of central directory
        self._comment = endrec[_ECD_COMMENT]  # archive comment

        # "concat" is zero, unless zip was concatenated to another file
        concat = endrec[_ECD_LOCATION] - size_cd - offset_cd
        if endrec[_ECD_SIGNATURE] == stringEndArchive64:
            # If Zip64 extension structures are present, account for them
            concat -= (sizeEndCentDir64 + sizeEndCentDir64Locator)

        if self.debug > 2:
            inferred = concat + offset_cd
            print("given, inferred, offset", offset_cd, inferred, concat)
        # self.start_dir:  Position of start of central directory
        self.start_dir = offset_cd + concat
        fp.seek(self.start_dir, 0)
        data = fp.read(size_cd)
        fp = io.BytesIO(data)
        total = 0
        while total < size_cd:
            centdir = fp.read(sizeCentralDir)
            if len(centdir) != sizeCentralDir:
                raise BadZipFile("Truncated central directory")
            centdir = struct.unpack(structCentralDir, centdir)
            if centdir[_CD_SIGNATURE] != stringCentralDir:
                raise BadZipFile("Bad magic number for central directory")
            if self.debug > 2:
                print(centdir)
            filename = fp.read(centdir[_CD_FILENAME_LENGTH])
            flags = centdir[5]
            if flags & 0x800:
                # UTF-8 file names extension
                filename = filename.decode('utf-8')
            else:
                # Historical ZIP filename encoding
                filename = filename.decode('gb18030')
            # Create ZipInfo instance to store file information
            x = ZipInfo(filename)
            x.extra = fp.read(centdir[_CD_EXTRA_FIELD_LENGTH])
            x.comment = fp.read(centdir[_CD_COMMENT_LENGTH])
            x.header_offset = centdir[_CD_LOCAL_HEADER_OFFSET]
            (x.create_version, x.create_system, x.extract_version, x.reserved,
             x.flag_bits, x.compress_type, t, d,
             x.CRC, x.compress_size, x.file_size) = centdir[1:12]
            if x.extract_version > MAX_EXTRACT_VERSION:
                raise NotImplementedError("zip file version %.1f" %
                                          (x.extract_version / 10))
            x.volume, x.internal_attr, x.external_attr = centdir[15:18]
            # Convert date/time code to (year, month, day, hour, min, sec)
            x._raw_time = t
            x.date_time = ((d >> 9) + 1980, (d >> 5) & 0xF, d & 0x1F,
                           t >> 11, (t >> 5) & 0x3F, (t & 0x1F) * 2)

            x._decodeExtra()
            x.header_offset = x.header_offset + concat
            self.filelist.append(x)
            self.NameToInfo[x.filename] = x

            # update total bytes read from central directory
            total = (total + sizeCentralDir + centdir[_CD_FILENAME_LENGTH]
                     + centdir[_CD_EXTRA_FIELD_LENGTH]
                     + centdir[_CD_COMMENT_LENGTH])

            if self.debug > 2:
                print("total", total)

    def open(self, name, mode="r", pwd=None, *, force_zip64=False):
        """Return file-like object for 'name'.

        name is a string for the file name within the ZIP file, or a ZipInfo
        object.

        mode should be 'r' to read a file already in the ZIP file, or 'w' to
        write to a file newly added to the archive.

        pwd is the password to decrypt files (only used for reading).

        When writing, if the file size is not known in advance but may exceed
        2 GiB, pass force_zip64 to use the ZIP64 format, which can handle large
        files.  If the size is known in advance, it is best to pass a ZipInfo
        instance for name, with zinfo.file_size set.
        """
        if mode not in {"r", "w"}:
            raise ValueError('open() requires mode "r" or "w"')
        if pwd and not isinstance(pwd, bytes):
            raise TypeError("pwd: expected bytes, got %s" % type(pwd).__name__)
        if pwd and (mode == "w"):
            raise ValueError("pwd is only supported for reading files")
        if not self.fp:
            raise ValueError(
                "Attempt to use ZIP archive that was already closed")

        # Make sure we have an info object
        if isinstance(name, ZipInfo):
            # 'name' is already an info object
            zinfo = name
        elif mode == 'w':
            zinfo = ZipInfo(name)
            zinfo.compress_type = self.compression
            zinfo._compresslevel = self.compresslevel
        else:
            # Get info object for name
            zinfo = self.getinfo(name)

        if mode == 'w':
            return self._open_to_write(zinfo, force_zip64=force_zip64)

        if self._writing:
            raise ValueError("Can't read from the ZIP file while there "
                             "is an open writing handle on it. "
                             "Close the writing handle before trying to read.")

        # Open for reading:
        self._fileRefCnt += 1
        zef_file = _SharedFile(self.fp, zinfo.header_offset,
                               self._fpclose, self._lock, lambda: self._writing)
        try:
            # Skip the file header:
            fheader = zef_file.read(sizeFileHeader)
            if len(fheader) != sizeFileHeader:
                raise BadZipFile("Truncated file header")
            fheader = struct.unpack(structFileHeader, fheader)
            if fheader[_FH_SIGNATURE] != stringFileHeader:
                raise BadZipFile("Bad magic number for file header")

            fname = zef_file.read(fheader[_FH_FILENAME_LENGTH])
            if fheader[_FH_EXTRA_FIELD_LENGTH]:
                zef_file.read(fheader[_FH_EXTRA_FIELD_LENGTH])
            if zinfo.flag_bits & 0x20:
                # Zip 2.7: compressed patched data
                raise NotImplementedError("compressed patched data (flag bit 5)")

            if zinfo.flag_bits & 0x40:
                # strong encryption
                raise NotImplementedError("strong encryption (flag bit 6)")

            if fheader[_FH_GENERAL_PURPOSE_FLAG_BITS] & 0x800:
                # UTF-8 filename
                fname_str = fname.decode("utf-8")
            else:
                fname_str = fname.decode("gb18030")

            if fname_str != zinfo.orig_filename:
                raise BadZipFile(
                    'File name in directory %r and header %r differ.'
                    % (zinfo.orig_filename, fname))

            # check for encrypted flag & handle password
            is_encrypted = zinfo.flag_bits & 0x1
            if is_encrypted:
                if not pwd:
                    pwd = self.pwd
                if not pwd:
                    raise RuntimeError("File %r is encrypted, password "
                                       "required for extraction" % name)
            else:
                pwd = None
            return ZipExtFile(zef_file, mode, zinfo, pwd, True)
        except Exception as e:
            zef_file.close()
            raise e


def iter_zip_file(
        file_path, unzip_path='', exclude=None, include=None,
        exclude_path=None, include_path=None, new_file=False, pwd=None):
    """unzip zip file"""
    if exclude is None:
        exclude = []
    if include is None:
        include = []
    if exclude_path is None:
        exclude_path = []
    if include_path is None:
        include_path = []
    zfile = GBKZipFile(file_path)
    if pwd:
        if isinstance(pwd, str):
            pwd = pwd.encode()
        zfile.setpassword(pwd)
    unzip_path = unzip_path or (file_path + ".tmp_unzip_files")

    logging.info(f'开始解压文件[{os.path.getsize(file_path) / 1024 / 1024:.3f}MB]：{file_path} ---> {unzip_path}')
    if os.path.exists(unzip_path):
        if new_file:
            remove_folder(unzip_path)
        else:
            raise OSError(f"解压错误：已存在文件：{unzip_path}")
    if os.path.isdir(unzip_path):
        pass
    else:
        os.mkdir(unzip_path)
    namelist = zfile.infolist()
    cnt = 0
    cnt_size = 0
    cnt_unzip = 0
    for idx, f_info in enumerate(namelist):
        cnt = idx
        f_path = f_info.filename
        # f_path_o = f_info.filename
        # f_path = f_path_o.encode('cp437').decode('gb18030')
        root, fn = os.path.split(f_path)
        if include and not any([bool(include_str in fn) for include_str in include]):
            continue
        if exclude and any([bool(exclude_str in fn) for exclude_str in exclude]):
            continue
        if include_path and not any([bool(include_str in f_path) for include_str in include_path]):
            continue
        if exclude_path and any([bool(exclude_str in f_path) for exclude_str in exclude_path]):
            continue

        f_size = f_info.file_size
        cnt_size += f_size

        logging.debug(f'解压到文件 [已处理{cnt}/{len(namelist)} 解压|{cnt_unzip}|个：{cnt_size / 1024 / 1024:.3f}MB]：{f_path} ')
        zfile.extract(f_path, unzip_path, pwd=pwd)
        # zfile.extract(f_path_o, unzip_path, pwd=pwd)
        # makedirs(os.path.join(unzip_path, f_path))
        # shutil.move(os.path.join(unzip_path, f_path_o), os.path.join(unzip_path, f_path))

        cnt_unzip += 1
        yield unzip_path, f_path
        os.remove(os.path.join(unzip_path, f_path))
    zfile.close()
    shutil.rmtree(unzip_path)
    logging.info(f'解压结束到文件夹 [{cnt_size / 1024 / 1024:.3f}MB] [已处理{cnt}|解压{cnt_unzip}个]：{unzip_path}')


def un_zip(
        file_path, unzip_path='', exclude=None, include=None,
        exclude_path=None, include_path=None, new_file=False, pwd=None):
    """unzip zip file"""
    if exclude is None:
        exclude = []
    if include is None:
        include = []
    if exclude_path is None:
        exclude_path = []
    if include_path is None:
        include_path = []
    zfile = GBKZipFile(file_path)
    if pwd:
        if isinstance(pwd, str):
            pwd = pwd.encode()
        zfile.setpassword(pwd)
    unzip_path = unzip_path or (file_path + ".tmp_unzip_files")

    logging.info(f'开始解压文件[{os.path.getsize(file_path) / 1024 / 1024:.3f}MB]：{file_path} ---> {unzip_path}')
    if os.path.exists(unzip_path):
        if new_file:
            remove_folder(unzip_path)
        else:
            raise OSError(f"解压错误：已存在文件：{unzip_path}")
    if os.path.isdir(unzip_path):
        pass
    else:
        os.mkdir(unzip_path)
    namelist = zfile.infolist()
    cnt = 0
    cnt_size = 0
    cnt_unzip = 0
    for idx, f_info in enumerate(namelist):
        cnt = idx
        f_path = f_info.filename
        # f_path_o = f_info.filename
        # f_path = f_path_o.encode('cp437').decode('gb18030')
        root, fn = os.path.split(f_path)
        if include and not any([bool(include_str in fn) for include_str in include]):
            continue
        if exclude and any([bool(exclude_str in fn) for exclude_str in exclude]):
            continue
        if include_path and not any([bool(include_str in f_path) for include_str in include_path]):
            continue
        if exclude_path and any([bool(exclude_str in f_path) for exclude_str in exclude_path]):
            continue

        f_size = f_info.file_size
        cnt_size += f_size

        logging.debug(f'解压到文件 [已处理{cnt}/{len(namelist)} 解压|{cnt_unzip}|个：{cnt_size / 1024 / 1024:.3f}MB]：{f_path} ')
        zfile.extract(f_path, unzip_path, pwd=pwd)

        # zfile.extract(f_path_o, unzip_path, pwd=pwd)
        # makedirs(os.path.join(unzip_path, f_path))
        # shutil.move(os.path.join(unzip_path, f_path_o), os.path.join(unzip_path, f_path))

        cnt_unzip += 1
    zfile.close()
    logging.info(f'解压结束到文件夹 [{cnt_size / 1024 / 1024:.3f}MB] [已处理{cnt}|解压{cnt_unzip}个]：{unzip_path}')
    return unzip_path


def iter_path(path, exclude=None, include=None, exclude_path=None, include_path=None, return_type=2, open_kwargs=None):
    if exclude is None:
        exclude = []
    if include is None:
        include = []
    if exclude_path is None:
        exclude_path = []
    if include_path is None:
        include_path = []
    if open_kwargs is None:
        open_kwargs = {}
    logging.info(f'开始遍历文件夹：{path}')
    cnt = 0
    cnt_size = 0
    cnt_iter = 0
    for root, fs, fns in os.walk(path):
        for fn in fns:
            cnt += 1
            if include and not any([bool(include_str in fn) for include_str in include]):
                continue
            if exclude and any([bool(exclude_str in fn) for exclude_str in exclude]):
                continue
            f_path = os.path.join(root, fn)
            if include_path and not any([bool(include_str in f_path) for include_str in include_path]):
                continue
            if exclude_path and any([bool(exclude_str in f_path) for exclude_str in exclude_path]):
                continue
            cnt_size += os.path.getsize(f_path)
            cnt_iter += 1
            logging.debug(f'遍历到文件：{f_path} [已处理{cnt_iter}|返回{cnt_iter}个：{cnt_size / 1024 / 1024:.3f}MB]')
            if return_type == 0:
                yield cnt, root, *os.path.splitext(fn)
            elif return_type == 1:
                yield cnt, root, fn
            elif return_type == 2:
                yield cnt, f_path
            elif return_type == 3:
                yield f_path
            elif return_type == 4:
                yield cnt, open(f_path, 'rb', **open_kwargs)
            elif return_type == 5:
                yield open(f_path, 'rb', **open_kwargs)
    logging.info(f'遍历文件结束 [{cnt_size / 1024 / 1024:.3f}MB] [已处理{cnt_iter}|返回{cnt_iter}个]：{path}')


def activate_debug_logger(level=logging.INFO):
    """Global logger used when running from command line."""
    logging.basicConfig(
        format='(%(levelname)s) %(message)s', level=level
    )


if __name__ == '__main__':
    activate_debug_logger()
    # zip_folder('./', './data/tmp.zip', root_folder='data')
    unzpath = un_zip("data.zip", include=['result.log'], new_file=True)
    list(iter_path(unzpath, return_type=3))



#!/usr/bin/env python3
# coding: utf-8
import traceback
import logging
from functools import wraps


def exception_wrapper(func):
    @wraps(func)
    def inner(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            func_info = f'{func.__code__.co_filename}:{func.__code__.co_firstlineno}:{func.__name__}()'
            logging.error('\n[%s]\n%r\n%r' % (func_info, args, kwargs),  e)

    return inner



import logging
import os

from unrar.cffi import rarfile
from d22d.utils.utils import remove_folder, makedirs


def un_rar(
        file_path, unzip_path='', exclude=None, include=None,
        exclude_path=None, include_path=None, new_file=False, pwd=None):
    """unzip zip file"""
    if exclude is None:
        exclude = []
    if include is None:
        include = []
    if exclude_path is None:
        exclude_path = []
    if include_path is None:
        include_path = []
    rfile = rarfile.RarFile(file_path)
    unzip_path = unzip_path or (file_path + "_files")

    logging.info(f'开始解压文件[{os.path.getsize(file_path) / 1024 / 1024:.3f}MB]：{file_path} ---> {unzip_path}')
    if os.path.exists(unzip_path):
        if new_file:
            remove_folder(unzip_path)
        else:
            raise OSError(f"解压错误：已存在文件：{unzip_path}")
    if os.path.isdir(unzip_path):
        pass
    else:
        os.mkdir(unzip_path)
    namelist = rfile.infolist()
    cnt = 0
    cnt_size = 0
    cnt_unzip = 0
    for idx, f_info in enumerate(namelist):
        cnt = idx
        f_path = f_info.filename
        # f_path_o = f_info.filename
        # f_path = f_path_o.encode('cp437').decode('gb18030')
        root, fn = os.path.split(f_path)
        if include and not any([bool(include_str in fn) for include_str in include]):
            continue
        if exclude and any([bool(exclude_str in fn) for exclude_str in exclude]):
            continue
        if include_path and not any([bool(include_str in f_path) for include_str in include_path]):
            continue
        if exclude_path and any([bool(exclude_str in f_path) for exclude_str in exclude_path]):
            continue

        f_size = f_info.file_size
        cnt_size += f_size

        logging.debug(f'解压到文件 [已处理{cnt}/{len(namelist)} 解压|{cnt_unzip}|个：{cnt_size / 1024 / 1024:.3f}MB]：{f_path} ')
        makedirs(os.path.join(unzip_path, f_path))
        if os.path.isdir(os.path.join(unzip_path, f_path)):
            continue
        with open(os.path.join(unzip_path, f_path), 'wb') as wf:
            wf.write(rfile.read(f_path))

        # zfile.extract(f_path_o, unzip_path, pwd=pwd)
        # makedirs(os.path.join(unzip_path, f_path))
        # shutil.move(os.path.join(unzip_path, f_path_o), os.path.join(unzip_path, f_path))

        cnt_unzip += 1

    logging.info(f'解压结束到文件夹 [{cnt_size / 1024 / 1024:.3f}MB] [已处理{cnt}|解压{cnt_unzip}个]：{unzip_path}')
    return unzip_path

# rar = rarfile.RarFile('360.rar')

# for filename in rar.namelist():
#     info = rar.getinfo(filename)
#     print("Reading {}, {}, {} bytes ({} bytes compressed)".format(info.filename, info.date_time, info.file_size, info.compress_size))
#     data = rar.read(filename)
#     print("\t{}...\n".format(data[:100]))



# /usr/bin/env python
# coding: utf-8

import logging
import logging.config
import datetime
import time
import sys
import os
import traceback
from enum import Enum
from tornado.log import LogFormatter as _LogFormatter

LOG_PATH = 'log'

def set_log_path(path):
    global LOG_PATH
    LOG_PATH = path


class ELogPriority(Enum):
    zdebug = 1
    zinfo = 2
    zwarn = 3
    zerror = 4
    zfatal = 5


try:
    import curses
except ImportError:
    curses = None

try:
    from concurrent_log_handler import ConcurrentRotatingFileHandler
except ImportError:
    from logging.handlers import RotatingFileHandler

    ConcurrentRotatingFileHandler = None


# tornado.log._stderr_supports_color = lambda: True


class LogFormatter(_LogFormatter, object):
    """Init tornado.log.LogFormatter from logging.config.fileConfig"""

    def __init__(self, fmt=None, datefmt=None, color=True, *args, **kwargs):
        if fmt is None:
            # fmt = _LogFormatter.DEFAULT_FORMAT
            fmt = '%(color)s[%(levelname)1.1s][%(asctime)s][%(name)s]' \
                  '[%(processName)s][%(threadName)s]' \
                  '[%(filename)s:%(lineno)d:%(funcName)s()]%(end_color)s' \
                  ' %(message)s'
            # fmt = '[%(asctime)s][%(name)s][%(levelname)s]'\
            #       '[%(processName)s][%(threadName)s]' \
            #       '[%(filename)s:%(lineno)d:%(funcName)s()]: %(message)s'
        super(LogFormatter, self).__init__(color=color, fmt=fmt, datefmt=datefmt, *args, **kwargs)

    def formatTime(self, record, datefmt="%Y-%m-%d %H:%M:%S.%f"):
        """
        Return the creation time of the specified LogRecord as formatted text.
        This method should be called from format() by a formatter which
        wants to make use of a formatted time. This method can be overridden
        in formatters to provide for any specific requirement, but the
        basic behaviour is as follows: if datefmt (a string) is specified,
        it is used with time.strftime() to format the creation time of the
        record. Otherwise, the ISO8601 format is used. The resulting
        string is returned. This function uses a user-configurable function
        to convert the creation time to a tuple. By default, time.localtime()
        is used; to change this for a particular formatter instance, set the
        'converter' attribute to a function with the same signature as
        time.localtime() or time.gmtime(). To change it for all formatters,
        for example if you want all logging times to be shown in GMT,
        set the 'converter' attribute in the Formatter class.
        """
        ct = self.converter(record.created)
        if datefmt:
            # s = time.strftime(datefmt, ct)
            s = str(datetime.datetime.now().strftime(datefmt))
        else:
            t = time.strftime("%Y-%m-%d %H:%M:%S", ct)
            s = "%s.%03d" % (t, record.msecs)
            # s = str(datetime.datetime.now().strftime(datefmt))
        return s


class SaveLogHandler(logging.Handler):
    """LogHandler that save records to a list"""

    def __init__(self, saveto=None, *args, **kwargs):
        self.saveto = saveto
        logging.Handler.__init__(self, *args, **kwargs)

    def emit(self, record):
        if self.saveto is not None:
            self.saveto.append(record)

    handle = emit


def enable_pretty_logging(logger=logging.getLogger()):
    channel = logging.StreamHandler()
    channel.setFormatter(LogFormatter())
    logger.addHandler(channel)


def get_realpath():
    return os.path.split(os.path.realpath(__file__))[0]


def get_logger_by_conf_file():
    while not os.path.exists('{}/log'.format(os.getcwd())):
        try:
            os.mkdir('{}/log'.format(os.getcwd()))
        except Exception as ex:
            import traceback
            err = traceback.format_exc()
            logging.error('{} {}'.format(ex, err))
            time.sleep(1)

    from logging.config import fileConfig
    fileConfig('%s/logging.conf' % get_realpath())
    return logging.getLogger(name='mainLogger')


def get_line_num_fast(filename):
    count = 0
    fp = open(filename, "rb")
    while 1:
        buffer = fp.read(1 * 1024 * 1024)
        if not buffer:
            break
        count += buffer.count(b'\n')
    logging.info("文件共统计到%d行" % count)
    fp.close()
    return count


class Logger:
    @classmethod
    def instance(cls):
        logging.info('日志目录：{}/log'.format(os.getcwd()))
        while not os.path.exists('{}/log'.format(os.getcwd())):
            try:
                os.mkdir('{}/log'.format(os.getcwd()))
            except Exception as ex:
                err = traceback.format_exc()
                logging.error(ex, err)
                logging.error('{} {}'.format(ex, err))
                time.sleep(1)

        from logging.config import fileConfig
        import sys
        sys.path.append(get_realpath())
        fileConfig('%s/logging.conf' % get_realpath())
        logger = logging.getLogger(name='mainLogger')

        # old_err_func = logger.error
        #
        # def _error(msg, ex=None, *args, **kwargs):
        #     if isinstance(msg, BaseException):
        #         temp_str = (ex.__str__() + "\n") or ""
        #         msg_f = f'{temp_str}{msg.__class__} {msg}\n{traceback.format_exc()}'
        #     elif isinstance(ex, BaseException):
        #         msg_f = f'{msg}\n{ex.__class__} {ex}\n{traceback.format_exc()}'
        #     else:
        #         msg_f = msg
        #     old_err_func(msg_f, *args, **kwargs)
        #
        # if logger.error is not _error:
        #     logger.error = _error
        return logger


log_conf_dict = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': "[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s",
            'datefmt': "%Y-%m-%d %H:%M:%S"
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
        "color": {
            "class": "d22d.utils.LogFormatter"
        },
    },
    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'color'
        },
        # 'file_root': {
        #     'level': 'DEBUG',
        #     'class': 'concurrent_log_handler.ConcurrentRotatingFileHandler' if
        #     ConcurrentRotatingFileHandler else 'logging.handlers.RotatingFileHandler',
        #     # 当达到10MB时分割日志
        #     'maxBytes': 1024 * 1024 * 500,
        #     # 最多保留50份文件
        #     'backupCount': 1,
        #     # If delay is true,
        #     # then file opening is deferred until the first call to emit().
        #     'delay': True,
        #     'filename': 'log/root.log',
        #     'formatter': 'color'
        # }
    },
    'loggers': {
        # 'root': {
        #     'handlers': ['file_root', 'console'],
        #     'level': 'DEBUG',
        # },
    }
}


def get_logger(name='mainLogger'):
    # global __g_logger
    # if name in __g_logger:
    #     return __g_logger[name]

    # print(logging.handlers)
    # from logging.config import fileConfig
    # fileConfig('%s/loggingnull.conf' % get_realpath())
    if not os.path.exists(LOG_PATH):
        os.makedirs(LOG_PATH)
    warn_l = ['WARNING', 'ERROR']
    all_l = ['DEBUG', 'INFO', 'WARNING', 'ERROR']
    for _name, levels in [
        ('requests', warn_l),
        ('urllib3', warn_l),
        ('elasticsearch', warn_l),
        ('sqlalchemy', warn_l),
        ('telethon', warn_l),
        ('kafka', warn_l),
        ('root', all_l),
        (name, all_l),
    ]:
        if _name in log_conf_dict:
            continue
        log_conf_dict['loggers'][_name] = {
            'handlers': ['console'] if name == _name else [],
            'level': levels[0],
        }
        for level in levels:
            log_conf_dict['handlers'][f'file_{_name}_{level}'] = {
                'level': level,
                'class': 'concurrent_log_handler.ConcurrentRotatingFileHandler' if
                ConcurrentRotatingFileHandler else 'logging.handlers.RotatingFileHandler',
                'maxBytes': 1024 * 1024 * 500,  # 当达到500MB时分割日志
                'backupCount': 1,  # 最多保留1份文件
                'delay': True,  # If delay is true, then file opening is deferred until the first call to emit().
                'filename': f'{LOG_PATH}/{_name}_{level.lower()}.log',
                'formatter': 'color'
            }
            log_conf_dict['loggers'][_name]['handlers'].append(f'file_{_name}_{level}')
        log_conf_dict['loggers'][_name]['handlers'] = list(set(log_conf_dict['loggers'][_name]['handlers']))
    logging.config.dictConfig(log_conf_dict)
    __logger = logging.getLogger(name=name)
    # __g_logger[name] = __logger
    # __logger.setLevel(logging.INFO)
    has_stdout = list(filter(lambda x: isinstance(x, logging.StreamHandler), __logger.handlers))
    if not has_stdout:
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(logging.Formatter(
            fmt='[%(levelname)s][%(asctime)s.%(msecs)03d]'
                '[%(processName)s:%(threadName)s:%(funcName)s:%(lineno)d] %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        ))
        __logger.addHandler(handler)

    host = os.environ.get('LOGSTASH_HOST')
    port = os.environ.get('LOGSTASH_PORT')
    # host = "192.168.0.73"
    # port = 32647
    from logstash import LogstashHandler

    has_handler = list(filter(lambda x: isinstance(x, LogstashHandler), __logger.handlers))
    if not has_handler:
        if host and port and (isinstance(port, int) or port.isdigit()):
            # from logstash_async.handler import AsynchronousLogstashHandler
            # logstash_handler = AsynchronousLogstashHandler(host, int(port), '')
            from logstash.formatter import LogstashFormatterBase

            def get_extra_fields(_, record):
                fields = {}
                msg = getattr(record, 'msg', None)
                if not isinstance(msg, str):
                    record.msg = str(msg)
                easy_types = (str, bool, dict, float, int, list, type(None))
                for key, value in record.__dict__.items():
                    if isinstance(value, easy_types):
                        fields[key] = value
                    else:
                        fields[key] = repr(value)

                return fields

            LogstashFormatterBase.get_extra_fields = get_extra_fields
            logstash_handler = LogstashHandler(host, int(port), 0)
            __logger.addHandler(logstash_handler)

    return __logger


if __name__ == '__main__':
    logz = Logger.instance()
    logz.debug('一个debug信息')
    logz.info('一个info信息')
    logz.warning('一个warning信息')
    try:
        i = [][1]
    except Exception as exxxx:
        logz.error('一个error信息', exxxx)



import asyncio
import time
import six
import wrapt
import logging
import os
import sys
import hashlib
from asyncio import sleep as asleep
from async_timeout import timeout as a_timeout


def get_file_name():
    sys_g = getattr(sys, '_getframe')
    _line = sys_g().f_back.f_lineno  # 调用此方法的代码的函数
    _file_name = sys_g(1).f_code.co_filename
    return _file_name


import threading

get_logger_lock = threading.Lock()
logger_task = logging.getLogger('d22d.task')
logger_atask = logging.getLogger('d22d.async_task')


# logging.basicConfig(
#     format='[%(levelname)1.1s][%(asctime)s][%(name)s][%(filename)s:%(lineno)d] %(message)s',
#     level=logging.INFO,
# )
#
#
# def get_logger(name):
#     formatter = logging.Formatter(
#                 fmt='''[%(levelname)1.1s][%(asctime)s][%(name)s] "%(pathname)s:%(lineno)d" %(message)s''')
#     sh = logging.StreamHandler()  # 创建屏幕输出handler
#     sh.setFormatter(formatter)   # 设置handler输出格式
#     sh.setLevel(logging.INFO)   # 设置handler输出等级
#     _log = logging.getLogger(name)   # 创建名为name值的logger
#     _log.setLevel(logging.INFO)
#
#     # 这句话是日志重复的关键
#     # 1.我们想为这个logger设置不一样的输出格式，所以添加了一个屏幕输出handler
#     # 2.没有设置logger的propagate值为False（默认是True），于是依据logging包官方的注释，日志记录会向上传播到他的父节点也就是root logger
#     # 3.当root logger 同时也被添加了屏幕输出handler的情况，日志就会输出第二次
#     # 4.解决方案有两个：一个是去掉root logger的屏幕输出handler，另一个是取消 子logger的 向上传播记录，也就是propagate值设为False
#     _log.addHandler(sh)
#
#     return _log
#
#
# log = get_logger('myselfApp')
# log.info('test')


def get_logger(name=None, formatter=None):
    # os.path.split(__file__)[-1].split(".")[0]

    if not formatter:
        formatter = logging.Formatter(
            fmt='''[%(levelname)1.1s][%(asctime)s][%(name)s] "%(pathname)s:%(lineno)d" %(message)s''')
        formatter_fh = logging.Formatter(fmt='''[%(levelname)1.1s][%(asctime)s][%(pathname)s:%(lineno)d] %(message)s''')
    else:
        formatter_fh = formatter

    if name and name != 'root':
        path = ''
        file_name = name
    elif name == 'root':
        path = ''
        file_name = name
        name = None
    else:
        sys_g = getattr(sys, '_getframe')
        _line = sys_g().f_back.f_lineno  # 调用此方法的代码的函数
        name = sys_g(1).f_code.co_filename
        # path = f'{os.path.dirname(name)}/'
        path = ''
        file_name = os.path.basename(name)
        name = file_name  # 加上这行让name短一点

    path_log = f'{os.getcwd()}/log/{path}'
    if not os.path.exists(path_log):
        os.makedirs(path_log)
    fh = logging.FileHandler(filename=f"{path_log}/{file_name}.log", mode="a", encoding="utf-8")
    fh_err = logging.FileHandler(filename=f"{path_log}/{file_name}_error.log", mode="a", encoding="utf-8")
    sh = logging.StreamHandler()

    fh.setFormatter(formatter_fh)
    fh_err.setFormatter(formatter_fh)
    sh.setFormatter(formatter)

    fh.setLevel(logging.INFO)
    fh_err.setLevel(logging.ERROR)
    sh.setLevel(logging.INFO)

    _log = logging.getLogger(name)
    if name is not None:
        _log.propagate = False

    get_logger_lock.acquire()
    for handler in _log.handlers:
        if isinstance(handler, logging.StreamHandler):
            _log.handlers.remove(handler)
    _log.addHandler(fh)
    _log.addHandler(fh_err)
    _log.addHandler(sh)
    # print(name, _log.handlers)
    get_logger_lock.release()

    _log.setLevel(logging.INFO)

    return _log


get_logger_and_add_file_stream_handler = get_logger
formatter_no_path = logging.Formatter(fmt='''[%(levelname)1.1s][%(asctime)s] %(message)s''')
# root_logger = get_logger('root')
# logger_timmer = get_logger('timmer', formatter=formatter_no_path)


def utf8(string):
    """
    Make sure string is utf8 encoded bytes.

    If parameter is a object, object.__str__ will been called before encode as bytes
    """
    if isinstance(string, six.text_type):
        return string.encode('utf8')
    elif isinstance(string, six.binary_type):
        return string
    else:
        return six.text_type(string).encode('utf8')


def get_md5(s):
    md5 = hashlib.md5(utf8(s)).hexdigest()
    return md5


def get_file_md5(file):
    md5 = hashlib.md5()
    f = open(file, 'rb')
    md5.update(f.read())
    f.close()
    return md5.hexdigest()


def async_task_auto_retry(
        warning_d=None,
        error_d=None,
        raise_e=None,
        time_sleep=5,
        max_retry=0,
        timeout=0,
        hard_timeout=0,
        # callback=None,
        # error_callback=None,
        finally_callback=None
):
    if not warning_d:
        warning_d = {}
    if not error_d:
        error_d = {}
    if not raise_e:
        raise_e = []

    """
      :param timeout: 任务超时时间
      :param hard_timeout: 任务超时时间退出
      :param max_retry: 任务最大重试次数
      :param time_sleep: 错误等待时长
      :param func: 任务函数
      :param args: args参数
      :param kwargs: kwargs参数
      :param warning_d: 报警异常类 以及 异常信息 kv字典
      {Exception: "这里是报警信息"}
      {Exception: lambda ex: f'这里是报警信息{ex.x}'}
      :param error_d: 错误异常类 以及 异常信息 kv字典
      {Exception: "这里是错误信息"}
      :param raise_e: 需要抛出异常的异常类
      [Exception]
      :return:
    """

    @wrapt.decorator
    async def wrapper(func, _instance, args, kwargs):
        nonlocal hard_timeout
        nonlocal timeout
        if callable(hard_timeout):
            _hard_timeout = hard_timeout(*args, **kwargs)
        else:
            _hard_timeout = hard_timeout
        if callable(timeout):
            _timeout = timeout(*args, **kwargs)
        else:
            _timeout = timeout

        if not isinstance(args, tuple):
            args = (args,)
        if not kwargs:
            kwargs = {}
        retry = 0
        time_start = time.time()
        func_info = f'{func.__code__.co_filename}:{func.__code__.co_firstlineno}:{func.__name__}()'
        args_info = '[ args: {} ]  |  [ kwargs: {} ]'.format(
            ', '.join(f'{arg}'[:50] for arg in args),
            ', '.join(f'{arg}={v}'[:50] for arg, v in kwargs.items()), )
        # if _instance:
        #     func = partial(func, _instance)
        while True:
            retry += 1
            try:
                logger_atask.debug(
                    f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info}' + ''
                    # f' {args} {kwargs}'
                )
                try:
                    if _hard_timeout:
                        async with a_timeout(_hard_timeout - int(time.time() - time_start)) as _:
                            result = await func(*args, **kwargs)
                            # print(_.expired)
                    else:
                        result = await func(*args, **kwargs)
                except asyncio.exceptions.TimeoutError:
                    e_t = TimeoutError(f'{func_info}超时 {_hard_timeout}s')
                    setattr(e_t, 'seconds', _hard_timeout)
                    raise e_t
                return result
            except Exception as ex:
                error = ex
                if ex.__class__ in warning_d:
                    ex_msg = warning_d[ex.__class__]
                    if ex_msg is not None:
                        if callable(ex_msg):
                            ex_msg = ex_msg(ex)
                        logger_atask.warning(
                            f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {args_info} {ex_msg}')
                if ex.__class__ in error_d:
                    ex_msg = error_d[ex.__class__]
                    if ex_msg is not None:
                        if callable(ex_msg):
                            ex_msg = ex_msg(ex)
                        logger_atask.error(
                            f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {args_info} {ex_msg}')
                else:
                    import traceback
                    logger_atask.error(
                        f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {args_info} '
                        f' {type(ex)} {str(ex)[:150]}'
                        f'\n{traceback.format_exc()}',
                        # f' {args} {kwargs}',
                    )
                if (raise_e is True) or (ex.__class__ in raise_e):
                    raise ex
            finally:
                if finally_callback is not None:
                    finally_callback(*args, **kwargs)
            if (
                    (max_retry and retry > max_retry) or
                    (_timeout and ((time.time() - time_start) > _timeout))
            ):
                raise error
            await asleep(time_sleep)

    return wrapper


def task_auto_retry(
        warning_d=None,
        error_d=None,
        raise_e=None,
        raise_all=None,
        time_sleep=60,
        max_retry=0,
        timeout=0,
):
    if not warning_d:
        warning_d = {}
    if not error_d:
        error_d = {}
    if not raise_e:
        raise_e = []
    """
      :param timeout: 任务超时时间
      :param max_retry: 任务最大重试次数
      :param time_sleep: 错误等待时长
      :param func: 任务函数
      :param args: args参数
      :param kwargs: kwargs参数
      :param warning_d: 报警异常类 以及 异常信息 kv字典
      {Exception: "这里是报警信息"}
      :param error_d: 错误异常类 以及 异常信息 kv字典
      {Exception: "这里是错误信息"}
      :param raise_e: 需要抛出异常的异常类
      [Exception]
      :return:
    """

    @wrapt.decorator
    def wrapper(func, _instance, args, kwargs):
        if not isinstance(args, tuple):
            args = (args,)
        if not kwargs:
            kwargs = {}
        retry = 0
        time_start = time.time()
        func_info = f'{func.__code__.co_filename}:{func.__code__.co_firstlineno}:{func.__name__}()'
        while True:
            retry += 1
            try:
                logger_task.debug(
                    f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info}' + ''
                    # f' {args} {kwargs}'
                )
                return func(*args, **kwargs)
            except Exception as ex:
                error = ex
                if ex.__class__ in warning_d:
                    error = ex
                    ex_msg = warning_d[ex.__class__]
                    if ex_msg is not None:
                        logger_task.warning(
                            f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
                if ex.__class__ in error_d:
                    error = ex
                    ex_msg = error_d[ex.__class__]
                    if ex_msg is not None:
                        logger_task.error(f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s] {func_info} {ex_msg}')
                else:
                    logger_task.error(
                        f'[RetryS:{retry:04d}:{time.time() - time_start:.2f}s]  {func_info}'
                        f' {type(ex)} {str(ex)[:150]}',
                        # f' {args} {kwargs}',
                    )
                if ex.__class__ in raise_e or raise_all:
                    raise ex
            if (
                    (max_retry and retry > max_retry) or
                    (timeout and ((time.time() - time_start) > timeout))
            ):
                raise error
            time.sleep(time_sleep)

    return wrapper


"""
    This class abstracts the well known startTime - time.time()
    it's a little redundant but it limits calls and allows to temporally pause
    time tracking
"""

import time


class Stopwatch(object):
    """
        defines the class that abstracts time tracking
    """

    def __init__(self):
        """
            initialize variables and control flags
        """
        self.start_time = -1.0
        self.elapsed_time = 0.0
        self.is_stopped = True
        self.is_paused = False
        self.reset_idx = 1
        self.start()

    def start(self):
        """
            sets flags and variables
        """
        if self.is_stopped and not self.is_paused:
            self.is_stopped = False
            self.elapsed_time = 0.0
        elif self.is_paused and not self.is_stopped:
            self.is_paused = False
        else:
            raise RuntimeError('Flags are unproperly set notify code author for him to revise logic')
        self.start_time = time.time()

    def stop(self):
        """
            stops stopwatch elapsed time gets stored but is lost upon restarting
        """

        if self.is_stopped:
            raise RuntimeError('Stopwatch is already stopped')
        self.elapsed_time += time.time() - self.start_time
        self.is_stopped = True
        self.is_paused = False

    def pause(self):
        """
            pauses the stopwatch without loosing elapsed time nor restarting it
        """
        if self.is_stopped:
            raise RuntimeError('Stopwatch is currently stopped, can\'t be paused')
        self.elapsed_time += time.time() - self.start_time
        self.is_paused = True

    def reset(self):
        """
            resets data without stopping the stopwatch itself
        """
        self.start_time = time.time()
        self.elapsed_time = 0
        self.reset_idx += 1

    def get_time(self, name=''):
        """
            returns the time elapsed since the stopwatch got started
            it's redundancy causes a small margin of error, please be weary
        """
        if not self.is_paused and not self.is_stopped:
            self.elapsed_time += time.time() - self.start_time
            self.start_time = time.time()
        print(f'[{self.reset_idx:02d}:{name}]花费了{self.elapsed_time:.5f}s')
        return self.elapsed_time

    def get_time_and_reset(self, name=''):
        """
            returns the time elapsed since the stopwatch got started
            it's redundancy causes a small margin of error, please be weary
        """
        res = self.get_time(name)
        self.reset()
        return res



import json
import logging
import os
import time
import threading
import unicodedata

from datetime import datetime, timedelta

from sys import platform
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from d22d.utils.utils import time_stamp, path_without_leaf

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class FileSync:
    def __init__(self):
        self.directoriesToWatch = []
        self.dictOfWatchedDir = {}
        self.synchAtStartupBool = False
        self.workingDir = "/"
        self.startBackupTime = "02:00"
        self.backupDurationInHours = 1
        self.delayToSend = -1
        self.listOfFilesToBeSent = []
        self.respectBackupHours = False
        self.ftp = None

    def sync_at_startup(self):
        try:
            if isinstance(self.directoriesToWatch, list) and len(self.directoriesToWatch) > 0:
                for dest in self.directoriesToWatch:
                    if 'localDirPath' in dest and 'remoteDirPath' in dest:
                        logger.info("synchAtStartup: synchronizing " + dest['localDirPath'])
                        arr = os.listdir(dest['localDirPath'])
                        for file in arr:
                            if not os.path.isdir(file):
                                self.moveFTPFiles(self.ftp, dest['localDirPath'] + "/" + file,
                                             dest['remoteDirPath'])
        except Exception as inst:
            logger.error("synchAtStartup Error - " + time_stamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly

    def moveFTPFiles(self, *args, **kwargs):
        return NotImplementedError

    def elaborateAllChangedFiles(self):
        global listOfFilesToBeSent
        if self.send_data_check():
            while listOfFilesToBeSent:
                file_to_upload = listOfFilesToBeSent[0]
                # file_to_upload = listOfFilesToBeSent.pop(0)
                path_without_filename = path_without_leaf(file_to_upload)
                if path_without_filename in self.dictOfWatchedDir:
                    self.moveFTPFiles(file_to_upload, self.dictOfWatchedDir[path_without_filename])
                    # when the system has finished to upload the file, we remove it from the list
                    listOfFilesToBeSent.remove(listOfFilesToBeSent[0])
                # listOfFilesToBeSent.pop(-1)
        else:
            logger.info("Not sending data because there could be a running update")
        # re-instantiate the thread that should send data
        threading.Timer(self.delayToSend, self.elaborateAllChangedFiles).start()

    def send_data_check(self):
        """
            we want to avoid sending data within an interval that goes
            among <backupDurationInHours> hour before the backup time
             and <backupDurationInHours> hour after the backup time
        """
        try:
            if self.respectBackupHours:
                if self.startBackupTime != "" and (":" in self.startBackupTime):
                    listOfParam = self.startBackupTime.split(":")
                    backupHour = listOfParam[0]
                    backupMinute = listOfParam[1]

                    minorCheck = datetime.today() - timedelta(hours=self.backupDurationInHours, minutes=0)
                    majorCheck = datetime.today() + timedelta(hours=self.backupDurationInHours, minutes=0)

                    date = datetime.now()
                    dateAndTimeOfBackup = date.replace(hour=int(backupHour), minute=int(backupMinute))

                    if dateAndTimeOfBackup > minorCheck and dateAndTimeOfBackup < majorCheck:
                        return False
                    else:
                        return True
                else:
                    # if anything goes wrong we return true
                    return True
            else:
                return True

        except Exception as inst:
            # if anything goes wrong we return true
            return True


class Handler(FileSystemEventHandler):
    def __init__(self, file_sync: FileSync, *args, **kwargs):
        self.file_sync = file_sync
        super().__init__(*args, **kwargs)

    def on_any_event(self, event):
        result = None
        try:
            if event.is_directory:
                # do nothing
                result = None
            elif event.event_type == 'created':
                # Take any action here when a file is first created.
                # when the file is create the "modified" event is also generated
                # so we do not anything
                logger.info("Received created event - %s." % event.src_path)
                # if (event.src_path in dictOfWatchedDir):
                #    moveFTPFiles(ftpServerName, ftpU, ftpP, event.src_path, dictOfWatchedDir[event.src_path], useTLS)

            elif event.event_type == 'modified':
                file_to_upload = event.src_path
                # Taken any action here when a file is modified.
                logger.info("Received modified event - %s." % file_to_upload)

                path_without_filename = path_without_leaf(file_to_upload)
                if path_without_filename in self.file_sync.dictOfWatchedDir:
                    if self.file_sync.delayToSend <= 0 and self.file_sync.send_data_check():
                        # directly send the file without any delay
                        self.file_sync.moveFTPFiles(file_to_upload, self.file_sync.dictOfWatchedDir[path_without_filename])
                    else:
                        # insert the name of the file in a list that will be elaborated only once every delayToSend seconds
                        if file_to_upload not in listOfFilesToBeSent:
                            listOfFilesToBeSent.append(file_to_upload)
            elif event.event_type == 'moved':
                # on linux, when a file is modified, the system create a temporal file (.goutputstream...)
                #  and them move it on the right one
                file_to_upload = event.dest_path

                # Taken any action here when a file is modified.
                logger.info("Received moved event - %s. - " % file_to_upload)

                path_without_filename = path_without_leaf(file_to_upload)
                if path_without_filename in self.file_sync.dictOfWatchedDir:
                    if self.file_sync.delayToSend <= 0 and self.file_sync.send_data_check():
                        # directly send the file without any delay
                        self.file_sync.moveFTPFiles(file_to_upload, self.file_sync.dictOfWatchedDir[path_without_filename])
                    else:
                        # insert the name of the file in a list that will be elaborated only once every delayToSend seconds
                        if (file_to_upload not in listOfFilesToBeSent):
                            listOfFilesToBeSent.append(file_to_upload)
                # path_without_filename = path_without_leaf(file_to_upload)
                # if (path_without_filename in dictOfWatchedDir):
                #     moveFTPFiles(ftpServerName, ftpU, ftpP, file_to_upload, dictOfWatchedDir[path_without_filename], useTLS)
            elif event.event_type == 'deleted':
                # do nothing
                result = None
        except Exception as inst:
            logger.error("on_any_event Error - " + time_stamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly


class Watcher(threading.Thread):
    # DIRECTORY_TO_WATCH = "/home/teo/Desktop/temp"

    def __init__(self, local_dir_path):
        try:
            threading.Thread.__init__(self)
            self.observer = Observer()
            self.localDirPath = local_dir_path
        except Exception as inst:
            logger.error("__init__ in Watcher Error for " + self.localDirPath + " - " + time_stamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly

    def run(self):
        try:
            event_handler = Handler(FileSync())
            self.observer.schedule(event_handler, self.localDirPath, recursive=True)
            self.observer.start()
            logger.debug("The thread to monitor " + self.localDirPath + " was run")

            try:
                while True:
                    time.sleep(5)
            except:
                self.observer.stop()
                logger.error("An exception occured in the thread loop of " + self.localDirPath)
                # we want a service that is always run
                w = Watcher(self.localDirPath)
                logger.info("Restart the thread to observe changes in files in" + self.localDirPath)
                w.start()

            logger.info("Join the thread for " + self.localDirPath)
            self.observer.join()

        except Exception as inst:
            logger.error("run in Watcher Error for " + self.localDirPath + " - " + time_stamp())
            logger.error(type(inst))  # the exception instance
            logger.error(inst.args)  # arguments stored in .args
            logger.error(inst)  # __str__ allows args to be printed directly


def move_file():
    conf = {
        "configuration": {
            "directoriesToWatch": [
                {
                    "localDirPath": "/home/user/temp",
                    "remoteDirPath": "temp1"
                },
                {
                    "localDirPath": "/home/user/temp2",
                    "remoteDirPath": "temp2"
                }
            ],
            "synchAtStartup": True,
            "ftpServerName": "server.domain.com",
            "ftpUser": "ftp-user",
            "ftpPass": "ftpPass",
            "useTLS": True,
            "workingDir": "/home/ftpWorkingDir",
            "delayToSend": 20,
            "startBackupTime": "10:12",
            "backupDurationInHours": 1,
            "respectBackupHours": True
        }
    }
    try:

        logger.info("loading configuration set")
        # read configuration parameters from the config.json file
        conf_path = 'config.json'
        if os.path.exists(conf_path):
            with open(conf_path) as json_data_file:
                data = json.load(json_data_file)
        if 'configuration' in data:
            conf = data['configuration']
        if 'directoriesToWatch' in conf:
            directoriesToWatch = conf['directoriesToWatch']
        if 'ftpServerName' in conf:
            ftpServerName = conf['ftpServerName']
        if 'ftpUser' in conf:
            ftpU = conf['ftpUser']
        if 'ftpPass' in conf:
            ftpP = conf['ftpPass']
        if 'useTLS' in conf:
            useTLS = conf['useTLS']
        if 'synchAtStartup' in conf:
            synchAtStartupBool = conf['synchAtStartup']
        if 'workingDir' in conf:
            workingDir = conf['workingDir']
        if 'startBackupTime' in conf:
            startBackupTime = conf['startBackupTime']
        if 'backupDurationInHours' in conf:
            backupDurationInHours = conf['backupDurationInHours']
        if 'respectBackupHours' in conf:
            respectBackupHours = conf['respectBackupHours']
        if (not respectBackupHours):
            logger.info("The script will NOT respect backup hours (so it will upload file also during backup hours)")

        # check if the folder exists, otherwise, create it
        try:
            if not os.path.exists(workingDir):
                os.makedirs(workingDir)
        except Exception as inst:
            if platform == "win32":
                workingDir = '/'
            else:
                workingDir = 'C:\\'
        try:
            if 'delayToSend' in conf:
                if (isinstance(conf['delayToSend'], int)):
                    delayToSend = conf['delayToSend']
                else:
                    delayToSend = int(conf['delayToSend'], 10)
        except Exception as inst:
            delayToSend = -1

        wrongPath = False

        logger.info("check correctness of links")
        if platform == "win32":
            # Windows...
            if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
                for dest in directoriesToWatch:
                    if ('/' in dest['localDirPath']):
                        wrongPath = True
                        logger.error("In the configuraton file you set a path (" + dest[
                            'localDirPath'] + ") as a linux-like path but you are under linux")
            # same check for the workingDir variable
            if ('/' in workingDir):
                wrongPath = True
                logger.error(
                    "In the configuraton file you set a path (" + workingDir + ") as a linux-like path but you are under linux")
        elif platform == "linux" or platform == "linux2" or platform == "darwin":
            # linux or OSX
            if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
                for dest in directoriesToWatch:
                    if ('\\' in dest['localDirPath']):
                        wrongPath = True
                        logger.error("In the configuraton file you set a path (" + dest[
                            'localDirPath'] + ") as a windows-like path but you are under linux")

            # same check for the workingDir variable
            if ('\\' in workingDir):
                wrongPath = True
                logger.error(
                    "In the configuraton file you set a path (" + workingDir + ") as a windows-like path but you are under linux")
        if (wrongPath):
            logger.error("Exiting due to errors in the path (e.g., we are under Windows but the user set a linux path)")
            exit()
        else:

            if synchAtStartupBool:
                logger.info("synchronizing files at startup")
                sync_at_startup()
            if (isinstance(directoriesToWatch, list) and len(directoriesToWatch) > 0):
                for dest in directoriesToWatch:
                    if 'localDirPath' in dest and 'remoteDirPath' in dest:
                        localDirPath = unicodedata.normalize('NFKD', dest['localDirPath']).encode('ascii', 'ignore')
                        remoteDirPath = unicodedata.normalize('NFKD', dest['remoteDirPath']).encode('ascii',
                                                                                                    'ignore')
                        dictOfWatchedDir[localDirPath] = remoteDirPath

                        w = Watcher(localDirPath)
                        logger.info("Start the thread to observe changes in files in" + localDirPath)
                        w.start()

            if (delayToSend > 0):
                threading.Timer(delayToSend, elaborateAllChangedFiles).start()
    except Exception as inst:
        logger.error("main Error - " + time_stamp())
        logger.error(type(inst))  # the exception instance
        logger.error(inst.args)  # arguments stored in .args
        logger.error(inst)  # __str__ allows args to be printed directly



import time


def str_to_int_time(string):
    return time.mktime(time.strptime(string, '%Y-%m-%d %H:%M:%S.%f')) - 3600 * 8


def int_to_str_time_gmtime(timeint):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(timeint))


def int_to_str_time_localtime(timeint):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(timeint))


def int_to_str_time_second(timeint):
    return time.strftime('%H:%M:%S', time.gmtime(timeint / 1000))


def get_time_day():
    return time.strftime("%Y-%m-%d", time.localtime())


def get_time_day_all():
    return time.strftime("%Y_%m_%d__%H_%M_%S", time.localtime())


def custom_time(timestamp):
    time_local = time.localtime(timestamp)
    dt = time.strftime("%Y-%m-%d", time_local)
    return dt



import base64
import html
import typing
import logging

import Crypto.Signature.PKCS1_v1_5 as sign_PKCS1_v1_5  # 用于签名/验签

from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5  # 用于加密
from Crypto import Random
from Crypto.Hash import SHA256

logger = logging.getLogger('ras')

"""
PKCS#1、PKCS#5、PKCS#7区别
PKCS5：PKCS5是8字节填充的，即填充一定数量的内容，使得成为8的整数倍，而填充的内容取决于需要填充的数目。例如，串0x56在经过PKCS5填充之后会成为0x56 0x07 0x07 0x07 0x07 0x07 0x07 0x07因为需要填充7字节，因此填充的内容就是7。当然特殊情况下，如果已经满足了8的整倍数，按照PKCS5的规则，仍然需要在尾部填充8个字节，并且内容是0x08,目的是为了加解密时统一处理填充。
PKCS7：PKCS7与PKCS5的区别在于PKCS5只填充到8字节，而PKCS7可以在1-255之间任意填充。
PKCS1：在进行RSA运算时需要将源数据D转化为Encryption block（EB）。其中pkcs1padding V1.5的填充模式按照以下方式进行
(1) EB = 00+BT+PS+00+D
EB：为填充后的16进制加密数据块，长度为1024/8 = 128字节（密钥长度1024位的情况下）
00：开头为00，是一个保留位
BT：用一个字节表示，在目前的版本上，有三个值00、01、02，如果使用公钥操作，BT为02，如果用私钥操作则可能为00或01
PS：填充位，PS = k － 3 － D 个字节，k表示密钥的字节长度，如果我们用1024bit的RSA密钥，k=1024/8=128字节，D表示明文数据D的字节长度，如果BT为00，则PS全部为00，如果BT为01，则PS全部为FF，如果BT为02，PS为随机产生的非0x00的字节数据。
00：在源数据D前一个字节用00表示
D：实际源数据
公式(1)整个EB的长度等于密钥的长度。
注意：对于BT为00的，数据D中的数据就不能以00字节开头，要不然会有歧义，因为这时候你PS填充的也是00，就分不清哪些是填充数据哪些是明文数据了，但如果你的明文数据就是以00字节开头怎么办呢？对于私钥操作，你可以把BT的值设为01，这时PS填充的FF，那么用00字节就可以区分填充数据和明文数据，对于公钥操作，填充的都是非00字节，也能够用00字节区分开。如果你使用私钥加密，建议你BT使用01，保证了安全性。
对于BT为02和01的，要保证PS至少要有八个字节长度
因为EB = 00+BT+PS+00+D = k
所以D <= k － 3 － 8，所以当我们使用128字节密钥对数据进行加密时，明文数据的长度不能超过128 － 11 = 117字节
当RSA要加密数据大于(k－11)字节时怎么办呢？把明文数据按照D的最大长度分块，然后逐块加密，最后把密文拼起来就行。
"""


class RSACtrl:
    def __init__(self, privatekey_path='tmp_privatekey.pem', publickey_path="tmp_publickey.pem", load_key_file=True):
        # 实现RSA 非对称加解密
        self.private_key = None  # 私钥
        self.public_key = None  # 公钥

        self.pri_obj = None  # 私钥obj
        self.pub_obj = None  # 公钥obj
        self.cert = None  # 证书

        self.private_key_max_handle_msg_len = 0
        self.public_key_max_handle_msg_len = 0

        self.privatekey_path = privatekey_path  # 私钥
        self.publickey_path = publickey_path  # 公钥
        if load_key_file:
            self.load_key_file()

    @staticmethod
    def generate_key(bit=2048):
        # 手动生成一个密钥对(项目中的密钥对一般由开发来生成)，生成密钥对的时候，可以指定生成的长度，一般推荐使用1024bit，
        # 1024bit的rsa公钥，最多只能加密117byte的数据，数据流超过这个数则需要对数据分段加密，
        # 目前1024bit长度的密钥已经被证明了不够安全，尽量使用2048bit长度的密钥，
        # 2048bit长度密钥最多能加密245byte长度的数据计算长度公式：密钥长度 / 8 - 11 = 最大加密量(单位bytes)下面生成一对2048bit的密钥：
        # x = RSA.generate(2048)
        x = RSA.generate(bit, Random.new().read)  # 也可以使用伪随机数来辅助生成
        privatekey = x.export_key()  # 私钥
        publickey = x.publickey().export_key()  # 公钥
        logger.debug(f"生成公私钥对：{type(privatekey)} {privatekey} {type(publickey)} {publickey}")
        return privatekey, publickey

    def generate_key_and_load(self, bit=2048):
        self.private_key, self.public_key = self.generate_key(bit)
        return self

    def import_public_key(self, publickey):
        self.public_key = publickey
        self.pub_obj = PKCS1_v1_5.new(RSA.importKey(self.public_key))
        self.public_key_max_handle_msg_len = getattr(self.pub_obj, '_key').size_in_bytes()
        return publickey

    def import_private_key(self, private_key, auto_public=True):
        pri_obj = RSA.importKey(private_key)
        self.cert = pri_obj.export_key("DER")  # 生成证书 -- 它和私钥是唯一对应的
        self.private_key = pri_obj.export_key()
        self.pri_obj = PKCS1_v1_5.new(pri_obj)
        # self.pri_obj = PKCS1_v1_5.new(pri_obj)
        self.private_key_max_handle_msg_len = getattr(self.pri_obj, '_key').size_in_bytes() - 11
        if auto_public:
            # 通过私钥生成公钥  (公钥不会变 -- 用于只知道私钥的情况)
            self.import_public_key(pri_obj.publickey().export_key())

        return pri_obj

    @staticmethod
    def get_data_form_file(path):
        with open(path, 'rb') as x:
            return x.read()

    def import_public_key_from_file(self, path="publickey.pem"):
        publickey = self.get_data_form_file(path)
        self.import_public_key(publickey)
        return publickey

    def import_private_key_obj_from_file(self, path='privatekey.pem', auto_public=True):
        # 从文件导入密钥
        private_key = self.get_data_form_file(path)
        self.privatekey_path = path
        return self.import_private_key(private_key, auto_public)

    def generate_key_and_save_file_and_load(self, bit=2048):
        """
        2048位RSA密钥生成需要2秒，4096需要15秒
        """
        self.generate_key_and_load(bit)
        self.save_key_file()
        self.load_key_file()
        return self

    def save_key_file(self):
        # 写入文件
        with open(self.privatekey_path, "wb") as x:
            x.write(self.private_key)
        with open(self.publickey_path, "wb") as x:
            x.write(self.public_key)
        return self

    def load_key_file(self):
        if self.publickey_path:
            publickey = self.import_public_key_from_file()
        else:
            publickey = None
        if self.privatekey_path:
            self.import_private_key_obj_from_file()
        if publickey != self.public_key:
            raise ValueError(f'公钥文件【{self.publickey_path}】和私钥文件【{self.privatekey_path}】不匹配')
        return self

    """
    ① 使用公钥 - 私钥对信息进行"加密" + “解密”
    作用：对信息进行公钥加密，私钥解密。 应用场景：
    A想要加密传输一份数据给B，担心使用对称加密算法易被他人破解（密钥只有一份，
    一旦泄露，则数据泄露），故使用非对称加密。
    信息接收方可以生成自己的秘钥对，即公私钥各一个，然后将公钥发给他人，
    私钥自己保留。

    A使用公钥加密数据，然后将加密后的密文发送给B，B再使用自己的私钥进行解密，
    这样即使A的公钥和密文均被第三方得到，
    第三方也要知晓私钥和加密算法才能解密密文，大大降低数据泄露风险。
    """

    def encrypt_with_rsa(self, plain_text):
        # 先公钥加密
        secret_byte_obj = self.pub_obj.encrypt(plain_text.encode())
        return secret_byte_obj

    def decrypt_with_rsa(self, secret_byte_obj):
        if isinstance(secret_byte_obj, str):
            secret_byte_obj = secret_byte_obj.encode()
        # 后私钥解密
        _byte_obj = self.pri_obj.decrypt(secret_byte_obj, Random.new().read)
        plain_text = _byte_obj.decode()
        return plain_text

    def encrypt_with_rsa_split_base64(self, msg, b64=True) -> str:
        """
        公钥加密 base64进行编码
        """
        # 分段加密
        encrypt_text = []
        if isinstance(msg, str):
            msg = msg.encode("utf-8")
        # 对数据进行分段加密
        for i in range(0, len(msg), self.private_key_max_handle_msg_len):
            cont = msg[i:i + self.private_key_max_handle_msg_len]
            encrypt_text.append(self.pub_obj.encrypt(cont))
        # 分段加密完进行拼接
        cipher_text = b''.join(encrypt_text)
        if b64:
            # base64进行编码
            cipher_text = base64.b64encode(cipher_text)
        return cipher_text.decode()

    def decrypt_with_rsa_split_base64(self, secret_byte_obj, encoding='utf-8', b64=True) -> str:
        if isinstance(secret_byte_obj, str):
            secret_byte_obj = secret_byte_obj.encode()
        if b64:
            secret_byte_obj = base64.b64decode(secret_byte_obj)

        # 分段加密
        encrypt_bytes = []
        msg_len = self.public_key_max_handle_msg_len
        # 对数据进行分段加密
        for i in range(0, len(secret_byte_obj), msg_len):
            cont = secret_byte_obj[i:i + msg_len]
            # 后私钥解密
            _byte_obj = self.pri_obj.decrypt(cont, Random.new().read)

            encrypt_bytes.append(_byte_obj)
        # 分段加密完进行拼接
        encrypt_byte = b''.join(encrypt_bytes)
        try:
            plain_text = encrypt_byte.decode(encoding)
        except Exception as ex:
            logger.warning(f'私钥解密 [ERROR] {ex}: {encoding.__repr__()}')
            plain_text = encrypt_byte.decode(encoding, errors='xmlcharrefreplace')
            plain_text = html.unescape(plain_text)
        return plain_text

    def to_sign_with_private_key(self, msg: typing.Union[str, bytes], b64=True) -> str:
        # 私钥签名
        signer_pri_obj = sign_PKCS1_v1_5.new(RSA.importKey(self.private_key))
        rand_hash = SHA256.new()
        if isinstance(msg, str):
            msg = msg.encode()
        rand_hash.update(msg)
        signature = signer_pri_obj.sign(rand_hash)
        if b64:
            signature = base64.b64encode(signature)
        return signature.decode()

    def to_verify_with_public_key(
            self, signature: typing.Union[str, bytes], msg: typing.Union[str, bytes], b64=True) -> bool:
        # 公钥验签
        verifier = sign_PKCS1_v1_5.new(RSA.importKey(self.public_key))
        _rand_hash = SHA256.new()
        if isinstance(msg, str):
            msg = msg.encode()
        _rand_hash.update(msg)
        try:
            if isinstance(signature, str):
                signature = signature.encode()
            if b64:
                signature = base64.b64decode(signature)
            verifier.verify(_rand_hash, signature)
            verify = True
        except ValueError as ex:
            if str(ex) == 'Invalid signature':
                verify = False
            else:
                raise ex
        return verify  # true / false

    def encode(self, msg: typing.Union[str, bytes], b64=True) -> str:
        """
        公钥加密
         str分段后 加密 join base64编码 返回str
        """
        return self.encrypt_with_rsa_split_base64(msg, b64=b64)

    def decode(self, secret: typing.Union[str, bytes], encoding='utf-8', b64=True) -> str:
        """
        私钥解密
         str base64解码 分段解密 join 返回str
        """
        return self.decrypt_with_rsa_split_base64(secret, encoding=encoding, b64=b64)

    def sign(self, msg: typing.Union[str, bytes], b64=True) -> str:
        """
        私钥签名
        """
        return self.to_sign_with_private_key(msg, b64=b64)

    def verify(self, signature: typing.Union[str, bytes], msg: typing.Union[str, bytes], b64=True) -> bool:
        """
        公钥验签
        """
        return self.to_verify_with_public_key(signature, msg, b64=b64)

    def check(self):
        """
        检测类函数方法以及公私密钥对可用性
        """
        self.verifier_with_signature(print)
        self.verifier_without_signature(print)

    def verifier_without_signature(self, logger_info: typing.Callable = logging.info):
        # 加解密验证
        text = "I love CA!"
        assert text == self.decrypt_with_rsa(self.encrypt_with_rsa(text))
        logger_info("rsa 加/解密验证 success！")

    def verifier_with_signature(self, logger_info: typing.Callable = logging.info):
        # 签名/验签
        text = "I love CA!"
        assert self.to_verify_with_public_key(self.to_sign_with_private_key(text), text)
        logger_info("rsa 签名/验签 success!")


if __name__ == '__main__':
    # __data = ''.join(str(i)+' ' for i in range(10000))
    __data = """
    ① 使用 公钥 - 私钥 对 信息进行 "加密" + “解密”
    作用：对信息进行公钥加密，私钥解密。 应用场景：
    A想要加密传输一份数据给B，担心使用对称加密算法易被他人破解（密钥只有一份，
    一旦泄露，则数据泄露），故使用非对称加密。
    信息接收方可以生成自己的秘钥对，即公私钥各一个，然后将公钥发给他人，
    私钥自己保留。

    A使用公钥加密数据，然后将加密后的密文发送给B，B再使用自己的私钥进行解密，
    这样即使A的公钥和密文均被第三方得到，
    第三方也要知晓私钥和加密算法才能解密密文，大大降低数据泄露风险。

    ② 使用 私钥 - 公钥 对 信息进行 "签名" + “验签”

    作用：对解密后的文件的完整性、真实性进行验证（繁琐但更加保险的做法，很少用到）
    应用场景： A有一私密文件欲加密后发送给B，又担心因各种原因导致B收到并解密后的文件并非完整、真实的原文件（可能被篡改或丢失一部分），所以A在发送前对原文件进行签名，将[签名和密文]一同发送给B让B收到后用做一下文件的 [解密 + 验签],
    均通过后-方可证明收到的原文件的真实性、完整性。

    如果是加密的同时又要签名，这个时候稍微有点复杂。
    1、发送者和接收者需要各持有一对公私钥，也就是4个钥匙。
    2、接收者的公私钥用于机密信息的加解密
    3、发送者的公私钥用于机密信息的签名/验签
    4、接收者和发送者都要提前将各自的[公钥]告知对方。
    """
    rsa_ctrl = RSACtrl()
    rsa_ctrl.check()

    __mi_wen = rsa_ctrl.encode(__data)
    print(__mi_wen)

    __ming_wen = rsa_ctrl.decode(__mi_wen)
    print(__ming_wen)

    __signature = rsa_ctrl.sign(__mi_wen)
    print(__signature)

    __is_ming_wen = rsa_ctrl.verify(__signature, __ming_wen)
    print(__is_ming_wen)



#!/usr/bin/env python3

#Dependencies to be installed:
#   pramiko
#   psutil (v3.4.2 for win XP)
#   pypiwin32 (219 for win XP)


import os
import platform
import sys

#Check if pip is installed
try:
    import pip
except ImportError:
    print('Please install pip. Failed to install dependencies.')
    input('Press enter to continue...')
    exit()

#if Windows
if(platform.system() == 'Windows'):
    print('Platform: Windows ' + platform.release())
    python_path = sys.executable
    if(python_path.split('\\')[-1:][0] == 'pythonw.exe'):
        print('Do not run this from IDLE. Double click the script to run it directly from python.exe')
        input('Press enter to continue...')
        exit()
    #if Windows 7/10
    if(platform.release() != 'XP' and platform.release() != '2003Server'):
        print('Upgrading pip...')
        os.system(python_path + ' -m pip install --upgrade pip')
        print('Installing pramiko...')
        os.system(python_path + ' -m pip install paramiko')
        print('Installing psutil...')
        os.system(python_path + ' -m pip install psutil')
        print('Installing pypiwin32...')
        os.system(python_path + ' -m pip install pypiwin32')
        input('Press enter to continue...')
    else:
    #if Windows XP
        print('Upgrading pip...')
        os.system(python_path + ' -m pip install --upgrade pip')
        print('Installing pramiko...')
        os.system(python_path + ' -m pip install paramiko')
        print('Installing psutil...')
        os.system(python_path + ' -m pip install psutil==3.4.2')
        print('Installing pypiwin32...')
        os.system(python_path + ' -m pip install pypiwin32==219')
        input('Press enter to continue...')       
#if Linux
elif(platform.system() == 'Linux'):
    print('Platform: Linux')
    print('Upgrading pip...')
    os.system('python3 -m pip install --upgrade pip --user')
    print('Installing pramiko...')
    os.system('python3 -m pip install paramiko --user')
    print('Installing psutil...')
    os.system('python3 -m pip install psutil --user')
    input('Press enter to continue...')
#if FreeBSD
elif(platform.system() == 'FreeBSD'):
    if(os.geteuid() != 0):
        print('You need to have root privileges to run this script.')
        input('Press enter to continue...')
    print('Platform: FreeBSD')
    print('Upgrading pip...')
    os.system('python3.6 -m pip install --upgrade pip')
    print('Installing tkinter...')
    os.system('pkg install py36-tkinter-3.6.5_6')
    print('Installing pramiko...')
    os.system('python3.6 -m pip install paramiko')
    print('Installing psutil...')
    os.system('python3.6 -m pip install psutil')
    input('Press enter to continue...')


# /usr/bin/env python
# coding: utf-8

import os
import sys
import json
import time
import datetime
import logging
from . import utils

if '-debug' in sys.argv:
    logger = utils.logger.get_logger('d22d')
else:
    logger = logging.getLogger('d22d')


def open_log():
    global logger
    logger = utils.logger.get_logger('d22d')


def format_value(data):
    if isinstance(data, float) and data % 1 == 0.0:
        data = int(data)
    elif isinstance(data, str):
        data = data.strip()
    return data


class Migration(object):
    def __init__(
            self, database_from, database_to, table_from=None, table_to=None,
            pks='', pkd=None, windows=10000, count_from=None, size=None, quchong=None,
            get_data_kwargs=None, save_data_kwargs=None):
        """
        转移整个数据库或者某一张表

        pks:
        为一个或多个主键的key值，使用","连接的一行字符串型参数，
        目前会使用在
            MysqlD.create_index, 创建表时设置主键
            ElasticSearchD.save_data，保存数据时依据data中某几个参数的值拼接 ”-“ 生成新值作为 _id


        pkd:
        在使用数据库到数据库，整个迁移的时候，设置每个表的pks，dict型
        例子：
            {
                "user": "name,age",
                "media": "user_id,media_id"
            }

        save_data_kwargs:
        设置save_data的参数，dict型
        {
            "data": action,  # 要存的数据list
            'index': table_to,  # 要存的表名
            'pks': pks,  # 上文中pks
            'batch_size': 1000,  # 每次批量存储的窗口大小，注意不是windows，须配合使用
            'mode': 'INSERT IGNORE'， # 目前是MysqlD的存储模式
        }
        """
        self.database_from = database_from
        self.database_to = database_to
        self.table_from = table_from
        self.table_to = table_to
        self.count_from = count_from
        self.size = size
        self.pks = pks
        self.pkd = pkd or {}
        self.windows = windows
        self.quchong = quchong or {}
        self.save_data_kwargs = save_data_kwargs or {}
        self.get_data_kwargs = get_data_kwargs or {}
        self.all_new_data_json_string = set({})

    def run(self):
        if self.table_from:
            table_to = self.table_to or self.table_from
            table_from = self.table_from
            self.run_one(table_from, table_to, self.pks)
        else:
            for table_from_raw in self.database_from.get_indexes():
                # table_from = table_from_raw
                # table_to = table_from_raw
                # if isinstance(self.database_from, utils.ElasticSearchD):
                #     table_from = {'index':table_from_raw}
                #
                # elif isinstance(self.database_from, utils.MySqlD):
                #     table_from = {'index':table_from_raw}
                #
                # if isinstance(self.database_to, utils.ElasticSearchD):
                #     table_to = {'index':table_from_raw}
                #
                # elif isinstance(self.database_to, utils.MySqlD):
                #     table_to = {'index':table_from_raw}

                self.run_one(table_from_raw, table_from_raw, self.pkd.get(table_from_raw, 'id'))
        del self.database_from
        del self.database_to

    def run_one(self, table_from, table_to, pks):
        gkws = {}
        gkws.update(self.get_data_kwargs)
        count = self.count_from or self.database_from.get_count(table_from, **gkws)
        table = self.database_from.get_data(table_from, **gkws)
        action = []
        time_start = time.time()
        first = False

        idx = 0
        for idx, d in enumerate(table):
            try:
                f_d = self.format_data(d)
                if not f_d:
                    continue
                if self.quchong:
                    if str(f_d) in self.all_new_data_json_string:
                        continue
                    else:
                        self.all_new_data_json_string.add(str(f_d))
            except Exception as e:
                logging.info(f'{self.database_from}/{table_from}:{idx} {e}')
                raise e
            action.append(f_d)
            if not first:
                first = True
                self.database_to.create_index(index=table_to, data=f_d, pks=pks)
            if (idx < 3) or (not idx % self.windows):
                # for k1,  k2, in zip(d.items(), f_d.items()):
                #     logger.info('{} -> {} | {}==>{}'.format(
                #
                #     ))
                # logger.info('{}\n|{}\n|\n|{}\n|{}\n'.format(f'{idx:-80}', d, f_d, f'{idx:-163}'))
                logger.info('[查看数据样例][{}] {}   -->   {}'.format(f'{idx:010}', d, f_d))

            if (self.size is not None) and (self.size <= idx + 1):
                break
            if len(action) >= self.windows:
                time_use = time.time() - time_start
                proc = (idx + 1) / count

                logger.info('[进度展示][{:012d}/{:012d}] {:0>8s}/{:0>8s}  ...{:.2f}%   {}/{} -> {}/{}'.format(
                    # self.database_to.get_count(table_to),
                    idx + 1, count, datetime.timedelta(seconds=time_use).__str__().split('.')[0],
                    datetime.timedelta(seconds=time_use / proc).__str__().split('.')[0],
                    proc * 100,
                    self.database_from, table_from,
                    self.database_to, table_to,
                ))
                kws = {"data": action, 'index': table_to, 'pks': pks, 'batch_size': 1000, 'mode': 'INSERT IGNORE'}
                kws.update(self.save_data_kwargs)
                utils.run_task_auto_retry(self.database_to.save_data, kwargs=kws)
                action = []
        if len(action):
            time_use = time.time() - time_start
            proc = 1
            logger.info('[进度展示][{:012d}/{:012d}] {:0>8s}/{:0>8s}  ...{:.2f}%   {}/{} -> {}/{}'.format(
                # self.database_to.get_count(table_to),
                idx + 1, count, datetime.timedelta(seconds=time_use).__str__().split('.')[0],
                datetime.timedelta(seconds=time_use / proc).__str__().split('.')[0],
                proc * 100,
                self.database_from, table_from,
                self.database_to, table_to,
            ))
            kws = {"data": action, 'index': table_to, 'pks': pks, 'batch_size': 1000, 'mode': 'INSERT IGNORE'}
            kws.update(self.save_data_kwargs)
            utils.run_task_auto_retry(self.database_to.save_data, kwargs=kws)
        # action = []

    @staticmethod
    def format_data(data):
        """
        修改table行数据再迁移到新的table

        :param data: dict table的行数据字典
        :return: dict 修改后table的行数据字典
        """
        return data


class Migration2DB(object):
    def __init__(
            self, database_from1, database_from2, database_to, table_from1, table_from2, table_to,
            migration_key1, migration_key2, pks='id', pkd=None, windows=1000,
            count_from1=None, count_from2=None, size=None, quchong=False, save_data_kwargs=None,
            get_data_kwargs1=None, get_data_kwargs2=None):
        self.database_from1 = database_from1
        self.database_from2 = database_from2
        self.database_to = database_to
        self.table_from1 = table_from1
        self.table_from2 = table_from2
        self.table_to = table_to
        self.migration_key1 = migration_key1
        self.migration_key2 = migration_key2 or migration_key1
        self.count_from1 = count_from1
        self.count_from2 = count_from2
        self.size = size
        self.pks = pks
        self.pkd = pkd or {}
        self.save_data_kwargs = save_data_kwargs or {}
        self.get_data_kwargs1 = get_data_kwargs1 or {}
        self.get_data_kwargs2 = get_data_kwargs2 or {}
        self.windows = windows
        self.quchong = quchong
        self.all_new_data_json_string = set({})

    def run(self):
        if self.table_from1:
            table_to = self.table_to or self.table_from1
            table_from = self.table_from1
            self.run_one(table_from, table_to, self.pks)
        else:
            for table_from_raw in self.database_from1.get_indexes():
                # table_from = table_from_raw
                # table_to = table_from_raw
                # if isinstance(self.database_from, utils.ElasticSearchD):
                #     table_from = {'index':table_from_raw}
                #
                # elif isinstance(self.database_from, utils.MySqlD):
                #     table_from = {'index':table_from_raw}
                #
                # if isinstance(self.database_to, utils.ElasticSearchD):
                #     table_to = {'index':table_from_raw}
                #
                # elif isinstance(self.database_to, utils.MySqlD):
                #     table_to = {'index':table_from_raw}

                self.run_one(table_from_raw, table_from_raw, self.pkd.get(table_from_raw, 'id'))

    def run_one(self, table_from1, table_to, pks):
        count1 = self.count_from1 or self.database_from1.get_count(table_from1)
        count2 = self.count_from2 or self.database_from1.get_count(table_from1)
        gkws1 = {}
        gkws1.update(self.get_data_kwargs1)
        gkws2 = {}
        gkws2.update(self.get_data_kwargs2)
        table = self.database_from1.get_data(table_from1, **gkws1)
        action = []
        time_start = time.time()
        table2d = {}
        for idx, d in enumerate(self.database_from2.get_data(self.table_from2, **gkws2)):
            if (idx < 3) or (not idx % self.windows):
                # logger.info('{}\n|{}'.format(f'{idx:080d}', d))
                logger.info('[数据样例展示][{}] {}'.format(f'{idx:010}', d))
                time_use = time.time() - time_start
                proc = (idx + 1) / count2

                logger.info('[进度展示][{:012d}/{:012d}] {:0>8s}/{:0>8s}  ...{:.2f}%   {}/{} Loading'.format(
                    idx + 1, count2, datetime.timedelta(seconds=time_use).__str__().split('.')[0],
                    datetime.timedelta(seconds=time_use / proc).__str__().split('.')[0],
                    proc * 100,
                    self.database_from2, self.table_from2,
                ))

            table2d[format_value(d[self.migration_key2]).__str__()] = {k: format(v) for k, v in d.items()}

        idx = 0
        for idx, d in enumerate(table):
            try:
                f_d = self.format_data(d, table2d.get(format_value(d.get(self.migration_key1, '')).__str__(), {}))

                if self.quchong:
                    if json.dumps(f_d) in self.all_new_data_json_string:
                        continue
                    else:
                        self.all_new_data_json_string.add(json.dumps(f_d))
            except Exception as e:
                logging.info(f'{self.database_from1}/{table_from1}:{idx} {e}')
                raise e
            action.append(f_d)
            if idx == 0:
                temp_d = {}
                for k, v in f_d.items():
                    temp_d[k] = v
                if table2d:
                    for k, v in table2d[list(table2d.keys())[0]].items():
                        temp_d[k] = v
                self.database_to.create_index(index=table_to, data=temp_d, pks=pks)
            if (idx < 3) or (not idx % self.windows):
                # for k1,  k2, in zip(d.items(), f_d.items()):
                #     logger.info('{} -> {} | {}==>{}'.format(
                #
                #     ))
                # logger.info('{}\n|{}\n|\n|{}\n|{}\n'.format(f'{idx:080d}', d, f_d, f'{idx:0163d}'))
                logger.info('[数据样例展示][{}] {}   -->   {}'.format(f'{idx:010}', d, f_d))
            if (self.size is not None) and (self.size <= idx + 1):
                break
            if len(action) >= self.windows:
                time_use = time.time() - time_start
                proc = (idx + 1) / count1

                logger.info('[进度展示][{:012d}/{:012d}] {:0>8s}/{:0>8s}  ...{:.2f}%   {}/{} -> {}/{}'.format(
                    # self.database_to.get_count(table_to),
                    idx + 1, count1, datetime.timedelta(seconds=time_use).__str__().split('.')[0],
                    datetime.timedelta(seconds=time_use / proc).__str__().split('.')[0],
                    proc * 100,
                    self.database_from1, table_from1,
                    self.database_to, table_to,
                ))
                kws = {"data": action, 'index': table_to, 'pks': pks, 'batch_size': 1000, 'mode': 'INSERT IGNORE'}
                kws.update(self.save_data_kwargs)
                utils.run_task_auto_retry(self.database_to.save_data, kwargs=kws)
                action = []
        if len(action):
            time_use = time.time() - time_start
            proc = 1
            logger.info('[进度展示][{:012d}/{:012d}] {:0>8s}/{:0>8s}  ...{:.2f}%   {}/{} -> {}/{}'.format(
                # self.database_to.get_count(table_to),
                idx + 1, count1, datetime.timedelta(seconds=time_use).__str__().split('.')[0],
                datetime.timedelta(seconds=time_use / proc).__str__().split('.')[0],
                proc * 100,
                self.database_from1, table_from1,
                self.database_to, table_to,
            ))
            kws = {"data": action, 'index': table_to, 'pks': pks, 'batch_size': 1000, 'mode': 'INSERT IGNORE'}
            kws.update(self.save_data_kwargs)
            utils.run_task_auto_retry(self.database_to.save_data, kwargs=kws)
        # action = []

    @staticmethod
    def format_data(data1, data2):
        new_data = data1
        for k, v in data2.items():
            new_data[k] = v

        return new_data



import pymysql
import threading
import wrapt
import psycopg2
import psycopg2.extras
from functools import update_wrapper, wraps
from dbutils.persistent_db import PersistentDB
from dbutils.pooled_db import PooledDB
import time
import json
import logging
from functools import partial

from d22d.utils.decorators import timmer, flyweight, where_is_it_called, print_hz


log = logging.getLogger(__name__)


class BaseController(object):

    def __init__(self, host, port, user, password, database):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.last_res = None
        self._conf = {}
        self._conn = None
        self._cur = None
        self._persist = None

    @property
    def persist(self):
        if not self._persist:
            self._persist = PooledDB(**self._conf)
        return self._persist

    @staticmethod
    def get_conn(persist):
        return persist.connection()

    @property
    def conn(self):
        if not self._conn or getattr(self._conn, "_closed", True):
            self._conn = self.get_conn(self.persist)
        return self._conn

    @staticmethod
    def get_cur(conn):
        return conn.cursor()

    @property
    def cur(self):
        if not self._cur or getattr(self._cur, "_closed", True):
            self._cur = self.get_cur(self.conn)
        return self._cur

    def execute(self, sql, params=None, return_type=2, res_windows=10000, commit=False):
        """

        :param sql:  sql命令
        :param params:  sql参数
        :param res_windows:  返回时游标窗口
        :param return_type:  返回类型 1：dict 2：list（dict） 3：迭代器
        :param commit:  自动提交
        :return:
        """
        # conn = self.conn
        conn = self.get_conn(self.persist)
        cur = self.get_cur(conn)
        cur.execute(sql, params)
        if return_type == 1:
            resl = {}
        elif return_type == 2:
            resl = []
        elif return_type == 4:
            resl = None
        else:
            raise ValueError(f"return_type {return_type} not in [1,2,3,4]")
        while res := cur.fetchmany(res_windows):
            for d in res:
                if return_type == 1:
                    resl = d
                    break
                elif return_type == 2:
                    resl.append(d)
                elif return_type == 4:
                    resl = list(d.values())
                    if len(resl):
                        resl = resl[0]
                    break
        cur.close()  # or del cur
        if commit:
            conn.commit()
        conn.close()  # or del db
        return resl

    def execute_iter(self, sql, params=None, res_windows=10000, commit=False):
        """

        :param sql:  sql命令
        :param params:  sql参数
        :param res_windows:  返回时游标窗口
        :param return_type:  返回类型 1：dict 2：list（dict） 3：迭代器
        :param commit:  自动提交
        :return:
        """
        # conn = self.conn
        conn = self.get_conn(self.persist)
        cur = self.get_cur(conn)
        cur.execute(sql, params)
        while res := cur.fetchmany(res_windows):
            for d in res:
                yield d
        cur.close()  # or del cur
        if commit:
            conn.commit()
        conn.close()  # or del db

    @where_is_it_called
    def execute_auto(self, sql, params=None, return_type=2, commit=False, *args, **kwargs):
        """
        return返回值
        return_type=1  第一个字典
        return_type=2  字典列表
        return_type=3  字典迭代器
        return_type=4  第一列第一行的值
        """
        # with decorators.ExceptionContextManager() as _:
        if return_type == 3:
            res = self.execute_iter(sql=sql, params=params, commit=commit, *args, **kwargs)
        else:
            res = self.execute(sql=sql, params=params, return_type=return_type, commit=commit, *args, **kwargs)
        return res

    def sql_select(self, table, where, res_key='*', limit=10, return_type=2, *args, **kwargs):
        return self.execute_auto(
            f'select {res_key} from {table} where {where} limit {limit}',
            return_type=return_type, *args, **kwargs)

    def sql_select_eq(self, table, where, res_key='*', limit=10, return_type=2, *args, **kwargs):
        keys, vals = format_sql_where(where)
        if isinstance(res_key, list):
            res_key = ', '.join(res_key)
        return self.execute_auto(
            f'select {res_key} from {table} where {keys} limit {limit}', params=vals,
            return_type=return_type, *args, **kwargs)

    def update_data(self, table_name, data):
        if data.get('timeupdate', False) is not False:
            data['timeupdate'] = time.time()
        aa = gen_update_sql_safe(table_name, data)
        return self.execute_auto(*aa, return_type=2, commit=True)

    def mysql_update(self, table_name, where, kwargs):
        ks, vs = format_sql_kv(kwargs)
        return self.execute_auto(
            'update {} set {} where {}'.format(
                table_name,
                ks,
                where
            ), params=vs, return_type=2, commit=True)

    def update_some_windows(self, table_name, update, where, windows=20000):
        flag = self.execute_auto(
            f'select count(0) as cnt from {table_name} '
            f'where {where}', return_type=4)
        while flag:
            print(f'重置数据库tasks剩余： {flag}')
            self.execute_auto(
                f'update {table_name} set {update} '
                f'where {where} limit {windows}', commit=True)

            if flag > windows:
                flag -= windows
            else:
                flag = self.execute_auto(
                    f'select count(0) as cnt from {table_name} '
                    f'where {where}',
                    return_type=4)

    @timmer()
    def update_data_by_pk(self, _table_name_, _pk_, _data_: list, **kwargs):
        kwargs["timeupdate"] = time.time()
        ks, vs = format_sql_kv(kwargs)
        return self.execute_auto(
            'update {} set {} where {} in ({})'.format(
                _table_name_,
                ks,
                _pk_,
                ', '.join([str(s[_pk_]) for s in _data_])
            ), params=vs, return_type=2, commit=True)

    get_one = partial(sql_select, return_type=1)
    get_some = partial(sql_select, return_type=2)


# 享元模式（Flyweight Pattern）主要用于减少创建对象的数量，以减少内存占用和提高性能。这种类型的设计模式属于结构型模式，它提供了减少对象数量从而改善应用所需的对象结构的方式。
@flyweight
class MYSQLController(BaseController):
    def __init__(self, host, port, user, password, database):
        super().__init__(host, port, user, password, database)
        self._conf = dict(
            # dbpai=pymysql,
            maxusage=1000,  # 一个连接最多复用次数，None为无限
            creator=pymysql,
            host=host,
            user=user,
            passwd=password,
            database=database if database else None,
            port=port,
            charset='utf8mb4',
            use_unicode=True,
            autocommit=False,
            # 流式、字典访问
            mincached=1,  # 初始化时，最小闲置数
            maxcached=10,  # 最大闲置数
            blocking=True,  # 无可用连接，等待
            # maxconnections=10, #创建连接池的最大数量
            # maxshared=20, #共享连接允许最大数量
            cursorclass=pymysql.cursors.SSDictCursor)


@flyweight
class PGController(BaseController):
    def __init__(self, host, port, user, password, database):
        super().__init__(host, port, user, password, database)
        self._conf = dict(
            maxusage=1000,
            creator=psycopg2,
            maxconnections=255,  # 连接池允许的最大连接数，0 和 None 表示不限制连接数
            mincached=1,  # 初始化时，链接池中至少创建的空闲的链接，0 表示不创建
            maxcached=32,  # 链接池中最多闲置的链接，0 和 None 不限制
            blocking=True,  # 连接池中如果没有可用连接后，是否阻塞等待。True，等待；False，不等待然后报错
            setsession=[],  # 开始会话前执行的命令列表
            host=host,
            user=user,
            password=password,
            database=database if database else None,
            port=port,
        )

    @staticmethod
    def get_cur(conn):
        return conn.cursor(
                name=str(time.time()), cursor_factory=psycopg2.extras.RealDictCursor, scrollable=True, withhold=True)

    def show_databases(self):
        return list(d.get('datname') for d in self.execute_auto('select datname from pg_database;', return_type=3))

    def show_tables(self, database=None):
        return list(d.get('table_name') for d in self.execute_auto(
            # f"SELECT table_name FROM information_schema.tables WHERE table_schema='{database or self.database}';", return_type=3))
            f"SELECT table_name FROM information_schema.tables WHERE table_catalog='{database or self.database}';", return_type=3))

    def describe_table_name(self, table_name):
        return list(d.get('column_name') for d in self.execute_auto(
            f"SELECT column_name FROM information_schema.columns WHERE table_name='{table_name}';", return_type=3))


def format_filed(filed):
    return ', '.join(f'{k}' for k in filed)


def format_sql_where(params):
    keys = []
    vals = []
    for k, v in params.items():
        keys.append(k)
        vals.append(v)
    return ' and '.join(f'{k}=%s' for k in keys), vals


def format_sql_kv(params):
    keys = []
    vals = []
    for k, v in params.items():
        keys.append(k)
        vals.append(v)
    return ', '.join(f'{k}=%s' for k in keys), vals


def gen_update_sql_safe(table_name, kvs):
    if len(kvs) == 0:
        return ''

    keys = []
    values = []
    for k, v in kvs.items():
        keys.append("`{}`".format(k))
        if isinstance(v, (tuple, set)):
            v = list(v)
        if isinstance(v, (dict, list)):
            v = json.dumps(v)
        values.append(v)
    sql = f'insert into {table_name}({",".join(keys)}) ' \
          f'values({",".join(["%s"] * len(values))})' \
          f'ON DUPLICATE KEY UPDATE {",".join("`%s`=%%s" % k for k in kvs.keys())}'

    return sql, values * 2



import xmlrpc.client
server = xmlrpc.client.Server('http://localhost:9001/RPC2')
print(server.system.methodHelp('supervisor.readProcessStdoutLog'))
print(server.supervisor.readProcessStdoutLog("lazyone", 100, 100))
print(server.supervisor.getState())
for p in server.supervisor.getAllProcessInfo():
    print(list(p.keys()))
    print(list(p.values()))
print(server.system.listMethods())
# server.system.methodHelp('supervisor.shutdown')